#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Session item for datasets and corresponding GUI."""

import operator

from numpy import sqrt, array, arange

from PyQt4.QtCore import SIGNAL, pyqtSignature as qtsig
from PyQt4.QtGui import QTabWidget, QWidget

from matplotlib.patches import Rectangle

from ufit.data.dataset import ScanData
from ufit.gui import logger
from ufit.gui.dataops import DataOps
from ufit.gui.session import session, SessionItem
from ufit.gui.scanitem import ScanDataItem
from ufit.gui.common import loadUi


class ImageDataItem(SessionItem):

    itemtype = 'image'

    def __init__(self, data):
        self.data = data
        SessionItem.__init__(self)

    def after_load(self):
        self.data.after_load()  # upgrade datastructures

    def __reduce__(self):
        return (self.__class__, (self.data, ))

    def create_panel(self, mainwindow, canvas):
        return ImageDataPanel(mainwindow, canvas, self)

    def create_multi_panel(self, mainwindow, canvas):
        return ImageMultiPanel(mainwindow, canvas)

    def update_htmldesc(self):
        title = self.data.title
        # XXX self.dataops.titleEdit.setText(title)
        self.title = title
        self.htmldesc = '<img src=":/image-sunset.png">&nbsp;&nbsp;' \
            '<big><b>%s</b></big>' % self.index + \
            (title and ' - %s' % title or '') + \
            (self.data.environment and
             '<br>%s' % ', '.join(self.data.environment) or '') + \
            ('<br><small>%s</small>' % '<br>'.join(self.data.sources))
        session.emit(SIGNAL('itemsUpdated'))

    def export_python(self, filename):
        pass

    def export_ascii(self, filename):
        with open(filename, 'wb') as fp:
            self.data.export_ascii(fp)

    def export_fits(self, filename):
        pass


class ImageDataPanel(QTabWidget):

    image_limits = None

    def __init__(self, parent, canvas, item):
        QTabWidget.__init__(self, parent)
        self.item = item
        self.dataops = DataOps(self)
        self._limits = None
        self.picker_widget = None

        self.canvas = canvas
        # XXX self.dataops.initialize(item)
        self.connect(self.dataops, SIGNAL('pickRequest'), self.set_picker)
        self.connect(self.dataops, SIGNAL('replotRequest'), self.plot)
        self.connect(self.dataops, SIGNAL('titleChanged'), self.item.update_htmldesc)
        self.addTab(self.dataops, 'Data operations')

    def set_picker(self, widget):
        self.picker_widget = widget

    def on_canvas_pick(self, event):
        if self.picker_widget:
            self.picker_widget.on_canvas_pick(event)

    def save_limits(self):
        # global limits for all images
        ImageDataPanel.image_limits = self.canvas.axes.get_xlim(), \
                                      self.canvas.axes.get_ylim()

    def get_saved_limits(self):
        return ImageDataPanel.image_limits

    def plot(self, limits=True, canvas=None):
        canvas = canvas or self.canvas
        plotter = canvas.plotter
        plotter.reset(limits)
        try:
            plotter.plot_image(self.item.data)
        except Exception:
            logger.exception('Error while plotting')
        else:
            canvas.draw()

    def export_ascii(self, filename):
        self.item.export_ascii(filename)

    def export_fits(self, filename):
        self.item.export_fits(filename)

    def export_python(self, filename):
        self.item.export_python(filename)


class ImageMultiPanel(QWidget):

    def __init__(self, parent, canvas):
        QWidget.__init__(self, parent)
        self.canvas = canvas
        self.boxes = []
        loadUi(self, 'imageops.ui')

    def initialize(self, items):
        self.items = [i for i in items if isinstance(i, ImageDataItem)]
        self.datas = [i.data for i in self.items]
        xparams = set()
        for data in self.datas:
            if not xparams:
                xparams.update(k for k in data.meta if isinstance(data.meta[k], float))
            xparams.intersection_update(k for k in data.meta
                                        if isinstance(data.meta[k], float))
        self.xparamBox.clear()
        xparams = ['image #'] + sorted(xparams)
        self.xparamBox.addItems(xparams)

    def plot(self, limits=True, canvas=None):
        canvas = canvas or self.canvas
        # XXX better title
        canvas.plotter.reset(ImageDataPanel.image_limits)
        sumdata = reduce(operator.add, self.datas[1:], self.datas[0])
        canvas.plotter.plot_image(sumdata)
        canvas.plotter.plot_finish(title='sum over %d images' % len(self.datas))
        for box in self.boxes:
            x1, y1, x2, y2 = box.x1Box.value(), box.y1Box.value(), \
                             box.x2Box.value(), box.y2Box.value()
            canvas.plotter.axes.add_patch(
                Rectangle((x1, y1), x2-x1, y2-y1, fill=False, color='yellow'))
        canvas.draw()

    def save_limits(self):
        ImageDataPanel.image_limits = self.canvas.axes.get_xlim(), \
                                      self.canvas.axes.get_ylim()

    @qtsig('')
    def on_addboxBtn_clicked(self):
        x1, x2 = map(int, self.canvas.axes.get_xlim())
        y1, y2 = map(int, self.canvas.axes.get_ylim())
        box = QWidget(self)
        loadUi(box, 'box.ui')
        # there's always a stretcher at the bottom
        self.boxLayout.insertWidget(self.boxLayout.count()-1, box)
        self.boxes.append(box)
        box.nameBox.setText('Box %d' % len(self.boxes))
        box.x1Box.setValue(x1)
        box.x2Box.setValue(x2)
        box.y1Box.setValue(y1)
        box.y2Box.setValue(y2)
        def boxchange(v):
            self.plot(False)
        def boxremove(box=box):
            index = self.boxes.index(box)
            del self.boxes[index]
            item = self.boxLayout.takeAt(index)
            item.widget().deleteLater()
            self.plot(False)
        self.connect(box.x1Box, SIGNAL('valueChanged(int)'), boxchange)
        self.connect(box.x2Box, SIGNAL('valueChanged(int)'), boxchange)
        self.connect(box.y1Box, SIGNAL('valueChanged(int)'), boxchange)
        self.connect(box.y2Box, SIGNAL('valueChanged(int)'), boxchange)
        self.connect(box.delBtn, SIGNAL('clicked()'), boxremove)
        self.plot(False)

    @qtsig('')
    def on_integrateBtn_clicked(self):
        xname = self.xparamBox.currentText()
        if xname == 'image #':
            xdata = arange(1, len(self.datas) + 1)
        else:
            xdata = array([data.meta[xname] for data in self.datas])
        boxnorm = self.boxNormBox.isChecked()
        for box in self.boxes:
            x1, y1, x2, y2 = box.x1Box.value(), box.y1Box.value(), \
                             box.x2Box.value(), box.y2Box.value()
            name = box.nameBox.text()
            ydata = array([data.arr[x1:x2, y1:y2].sum()
                           for data in self.datas])
            dydata = array([sqrt((data.darr[x1:x2, y1:y2]**2).sum())
                            for data in self.datas])
            yname = 'box counts'
            if boxnorm:
                factor = 1. / ((y2 - y1) * (x2 - x1))
                ydata *= factor
                dydata *= factor
                yname = 'box counts (norm)'
            scan = ScanData.from_arrays(name, xdata, ydata, dydata,
                                        xcol=xname, ycol=yname)
            session.add_item(ScanDataItem(scan))



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Model builder panel."""

import re

from PyQt4.QtCore import pyqtSignature as qtsig, SIGNAL
from PyQt4.QtGui import QWidget, QListWidgetItem, QDialogButtonBox, \
    QMessageBox, QInputDialog, QTextCursor, QDialog

from ufit.models import concrete_models, eval_model
from ufit.models.corr import Background
from ufit.models.peaks import GaussInt
from ufit.gui import logger
from ufit.gui.common import loadUi

ident_re = re.compile('[a-zA-Z][a-zA-Z0-9_]*$')


class ModelBuilder(QWidget):

    def __init__(self, parent):
        QWidget.__init__(self, parent)
        self.logger = logger.getChild('model')
        self.gauss_picking = 0
        self.gauss_peak_pos = 0, 0
        self.pick_model = None
        self.data = None
        self.last_model = None
        self.createUI()

    def createUI(self):
        loadUi(self, 'modelbuilder.ui')
        self.modeldefStacker.setCurrentIndex(0)
        self.model_dict = {}
        for model in concrete_models:
            QListWidgetItem(model.__name__, self.premodelsList)
            self.model_dict[model.__name__] = model
        self.buttonBox.addButton('Check', QDialogButtonBox.NoRole)
        self.buttonBox.addButton(QDialogButtonBox.Apply)

    def on_buttonBox_clicked(self, button):
        role = self.buttonBox.buttonRole(button)
        if role == QDialogButtonBox.ResetRole:
            self.modeldefEdit.setText('')
        elif role == QDialogButtonBox.NoRole:
            self.eval_model()
        else:  # "apply"
            self.eval_model(final=True)

    @qtsig('')
    def on_gaussOnlyBtn_clicked(self):
        if self.gauss_picking:
            self._finish_picking()
            return
        self.gaussOnlyBtn.setText('Back to full modeling mode')
        self.emit(SIGNAL('pickRequest'), self)
        self.gauss_picking = 1
        self.gauss_picked_points = []
        self.modeldefStacker.setCurrentIndex(1)
        self.pick_model = Background(bkgd=self.data.y.min())
        self.emit(SIGNAL('newModel'), self.pick_model, True, False)

    def on_canvas_pick(self, event):
        if not self.gauss_picking:
            return
        if hasattr(event, 'artist'):
            return
        if self.gauss_picking % 2 == 1:
            # first click, picked position
            self.gauss_peak_pos = event.xdata, event.ydata
        else:
            # second click, picked width
            pos = self.gauss_peak_pos[0]
            ampl = self.gauss_peak_pos[1] - self.data.y.min()
            fwhm = abs(pos - event.xdata) * 2
            self.pick_model += GaussInt('p%02d' % (self.gauss_picking/2),
                                        pos=pos, int=fwhm*ampl*2.5, fwhm=fwhm)
            self.emit(SIGNAL('newModel'), self.pick_model, True, False)
        self.gauss_picking += 1

    def _finish_picking(self):
        if not self.gauss_picking:
            return
        self.gauss_picking = None
        self.gaussOnlyBtn.setText('Gauss peaks only mode')
        self.modeldefStacker.setCurrentIndex(0)

    def on_premodelsList_currentItemChanged(self, current, previous):
        model = self.model_dict[str(current.text())]
        self.modelinfoLbl.setText(model.__doc__)

    def on_premodelsList_itemDoubleClicked(self, item):
        self.on_addmodelBtn_clicked()

    @qtsig('')
    def on_addmodelBtn_clicked(self):
        modelitem = self.premodelsList.currentItem()
        if not modelitem:
            return
        modelcls = str(modelitem.text())
        modelname = QInputDialog.getText(self, 'ufit', 'Please enter a name '
                                         'for the model part:')[0]
        if not modelname:
            return
        self.insert_model_code('%s(%r)' % (modelcls, str(modelname)))

    @qtsig('')
    def on_addCustomBtn_clicked(self):
        dlg = QDialog(self)
        loadUi(dlg, 'custommodel.ui')
        while 1:
            if dlg.exec_() != QDialog.Accepted:
                return
            modelname = str(dlg.nameBox.text())
            params = str(dlg.paramBox.text())
            value = str(dlg.valueEdit.toPlainText()).strip()
            if not ident_re.match(modelname):
                QMessageBox.warning(self, 'Error', 'Please enter a valid model '
                                    'name (must be a Python identifier using '
                                    'only alphabetic characters and digits).')
                continue
            if not params:
                QMessageBox.warning(self, 'Error', 'Please enter some parameters.')
                continue
            for param in params.split():
                if not ident_re.match(param):
                    QMessageBox.warning(
                        self, 'Error', 'Parameter name %s is not valid (must '
                        'be a Python identifier using only alphabetic '
                        'characters and digits).' % param)
                    params = None
                    break
            if not params:
                continue
            break
        self.insert_model_code('Custom(%r, %r, %r)' % (modelname, params, value))

    def insert_model_code(self, code):
        currentmodel = str(self.modeldefEdit.toPlainText())
        prefix = ''
        if currentmodel:
            prefix = ' + '
        tc = self.modeldefEdit.textCursor()
        tc.movePosition(QTextCursor.End)
        tc.insertText(prefix + code)

    def initialize(self, data, model):
        self.model = model
        self.data = data
        self.modeldefEdit.setText(model.get_description())

    def eval_model(self, final=False):
        modeldef = str(self.modeldefEdit.toPlainText()).replace('\n', ' ')
        if not modeldef:
            QMessageBox.information(self, 'Error', 'No model defined.')
            return
        try:
            model = eval_model(modeldef)
        except Exception, e:
            self.logger.exception('Could not evaluate model')
            QMessageBox.information(self, 'Error',
                                    'Could not evaluate model: %s' % e)
            return
        if final:
            self._finish_picking()
            self.last_model = model
            self.emit(SIGNAL('newModel'), model)
            self.emit(SIGNAL('closeRequest'))
        else:
            self.statusLbl.setText('Model definition is good.')



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Load routine for LLB TAS data (binary format)."""

import struct
from time import mktime

from numpy import array, sqrt

# format of the first 10 bytes of the header
DATEFMT = struct.Struct('<hhhhh')

# names the header floats starting at position 92; None means unmapped
# CAUTION: some of these names are only correct for 1T
FIELDS = \
    ['dm', 'da', 'etam', 'etaa', 'zme', 'zea', 'zad', 'he', 've', 'hd', 'vd'] + \
    [None] * 12 + ['h0', 'h1', 'h2', 'h3', 'v0', 'v1', 'v2', 'v3',
    'eta1', 'eta2', 'eta3'] + [None] * 12 + [
    'ax', 'ay', 'az', 'alfa', 'beta', 'gama', 'etas'] + [None] * 18 + [
    'ze1', 'ze2', 'za1', 'za2'] + [None] * 4 + ['zgi', 'zgs', None, 'zca']

# format of a single point (92 bytes long)
POINTFMT = struct.Struct('<hffffffffffffffhffffffff')
POINTFIELDS = ['qh', 'qk', 'ql', 'en', 'ki', 'm1', 'm2',
               'e1', 'e2', 'a1', 'a2', 'time', 'mon', 'counts', 'xx', 'T']

# format for 4F with double monochromator
POINTFMT_alt = struct.Struct('<hffffffffffffffffhffffff')
POINTFIELDS_alt = ['qh', 'qk', 'ql', 'en', 'ki', 'm1', 'm2', 'm3', 'm4',
                   'e1', 'e2', 'a1', 'a2', 'time', 'mon', 'counts', 'xx', 'T']

def check_data(fp):
    # the first 10 bytes are a packed date; check that for plausibility
    dates = fp.read(10)
    fp.seek(0, 0)
    if len(dates) < 10:
        return False
    d, m, y, hh, mm = DATEFMT.unpack(dates)
    return (1900 < y < 2100) and (1 <= m <= 12) and (1 <= d <= 31)


def guess_cols(colnames, coldata, meta):
    return meta['hkle_vary'], 'counts', None, 'mon'


def _float_fmt(x):
    if x == 0:
        return '0'
    else:
        return '%.3f' % x

def read_data(filename, fp):
    meta = {}
    dates = fp.read(10)
    d, m, y, hh, mm = DATEFMT.unpack(dates)
    meta['created'] = mktime((y, m, d, hh, mm, 0, 0, 0, 0))
    titre = fp.read(80).strip()
    fp.read(2)  # skip null bytes
    meta['title'] = titre
    headerfields = struct.unpack('<115f', fp.read(460))
    for i, name in enumerate(FIELDS):
        if name is None:
            continue
        meta[name] = headerfields[i]
    # reconstruct run command from header fields
    meta['subtitle'] = 'run q=(%s %s %s %s) dq=(%s %s %s %s) ' % \
                       tuple(map(_float_fmt, headerfields[92:100]))
    meta['subtitle'] += 'np=%d m=%d ' % headerfields[100:102]
    if headerfields[102] == 1:
        # const ki
        meta['subtitle'] += 'ki=%.3f' % headerfields[103]
    else:
        meta['subtitle'] += 'kf=%.3f' % headerfields[103]
    parr = []
    pointfmt = POINTFMT
    pointfields = POINTFIELDS
    for i, point in enumerate(iter(lambda: fp.read(POINTFMT.size), '')):
        if i == 0:
            unp = pointfmt.unpack(point)
            if unp[14] != 1:
                pointfmt = POINTFMT_alt
                pointfields = POINTFIELDS_alt
        coords = pointfmt.unpack(point)[:len(pointfields) + 1]
        parr.append(coords + (coords[1] + 0.5*coords[2],
                              0.5*sqrt(3)*coords[2]))
    parr = array(parr)[:,1:]
    for i, name in enumerate(pointfields):
        meta[name] = parr[:,i].mean()
    meta['environment'] = ['T = %.3f K' % meta['T']]
    meta['hkle'] = parr[:,0:4]
    deviations = array([cs.max() - cs.min() for cs in parr.T[0:4]])
    xg = pointfields[deviations.argmax()]
    meta['hkle_vary'] = xg
    return pointfields + ['qx', 'qy'], parr, meta



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Metadata view window."""

from PyQt4.QtCore import SIGNAL, QByteArray, Qt
from PyQt4.QtGui import QMainWindow, QTableWidgetItem, QMessageBox

from ufit.gui.common import loadUi, SettingGroup
from ufit.gui.session import session


class InspectorWindow(QMainWindow):
    def __init__(self, parent):
        self._updating = False
        self._data = None
        QMainWindow.__init__(self, parent)
        loadUi(self, 'inspector.ui')
        #layout = QVBoxLayout()
        #layout.setContentsMargins(0, 0, 0, 0)
        #layout.addWidget(self.canvas)
        self.sgroup = SettingGroup('inspector')
        self.tbl.verticalHeader().setDefaultSectionSize(
            self.tbl.verticalHeader().minimumSectionSize() + 2)

        with self.sgroup as settings:
            geometry = settings.value('geometry', QByteArray())
            self.restoreGeometry(geometry)
            windowstate = settings.value('windowstate', QByteArray())
            self.restoreState(windowstate)

    def setDataset(self, data):
        self.data = data
        self.dataName.setText('%s - %s' % (data.name, data.title))
        self._updating = True
        self.tbl.setRowCount(len(data.meta))
        for i, key in enumerate(sorted(data.meta, key=lambda n: n.lower())):
            key_item = QTableWidgetItem(key)
            key_item.setFlags(key_item.flags() & ~Qt.ItemIsEditable)
            self.tbl.setItem(i, 0, key_item)
            if key.startswith('col_'):
                value_item = QTableWidgetItem(str(data.meta[key]))
                value_item.setFlags(value_item.flags() & ~Qt.ItemIsEditable)
            else:
                value_item = QTableWidgetItem(repr(data.meta[key]))
            self.tbl.setItem(i, 1, value_item)
        self._updating = False

    def closeEvent(self, event):
        with self.sgroup as settings:
            settings.setValue('geometry', self.saveGeometry())
            settings.setValue('windowstate', self.saveState())
        self.emit(SIGNAL('closed'))
        return QMainWindow.closeEvent(self, event)

    def on_tbl_itemChanged(self, item):
        if self._updating:
            return
        try:
            new_value = eval(str(item.text()))
        except Exception:
            QMessageBox.error(self, 'Error',
                              'The new value is not a valid expression.')
            return
        else:
            key = str(self.tbl.item(item.row(), 0).text())
            self.data.meta[key] = new_value
        self.emit(SIGNAL('replotRequest'), None)
        session.set_dirty()



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Common GUI elements."""

import sys
from os import path
from cStringIO import StringIO

from PyQt4 import uic
from PyQt4.QtCore import SIGNAL, QSize, QSettings, Qt, QRectF, QByteArray
from PyQt4.QtGui import QLineEdit, QSizePolicy, QWidget, QIcon, QFileDialog, \
    QMessageBox, QPrinter, QPrintDialog, QPrintPreviewWidget, QPainter, QDialog
from PyQt4.QtSvg import QSvgRenderer

from matplotlib.backends.backend_qt4agg import \
    FigureCanvasQTAgg as FigureCanvas, NavigationToolbar2QT, FigureManagerQT
from matplotlib._pylab_helpers import Gcf
from matplotlib.colors import LogNorm
from matplotlib.figure import Figure
from matplotlib import pyplot
try:
    from matplotlib.backend_bases import key_press_handler
except ImportError:
    key_press_handler = None
from matplotlib.backends import backend_qt4
backend_qt4.figureoptions = None  # disable toolbar button that doesn't work
pyplot.rc('font', family='sans-serif')
pyplot.rc('font', **{'sans-serif': 'Sans Serif, Arial, Helvetica, '
                     'Lucida Grande, Bitstream Vera Sans'})

from ufit.gui import logger
from ufit.gui.session import session
from ufit.plotting import DataPlotter

uipath = path.dirname(__file__)

def loadUi(widget, uiname, subdir='ui'):
    uic.loadUi(path.join(uipath, subdir, uiname), widget)

def path_to_str(qstring):
    return unicode(qstring).encode(sys.getfilesystemencoding())

def str_to_path(string):
    return string.decode(sys.getfilesystemencoding())


class MPLCanvas(FigureCanvas):
    """Ultimately, this is a QWidget (as well as a FigureCanvasAgg, etc.)."""
    def __init__(self, parent, width=10, height=6, dpi=72, maincanvas=False):
        fig = Figure(figsize=(width, height), dpi=dpi)
        fig.set_facecolor('white')
        self.printer = None
        self.print_width = 0
        self.main = parent
        self.logz = False
        self.axes = fig.add_subplot(111)
        self.plotter = DataPlotter(self, self.axes)
        # make tight_layout do the right thing
        self.axes.set_xlabel('x')
        self.axes.set_ylabel('y')
        self.axes.set_title('(data title)\n(info)', size='medium')
        FigureCanvas.__init__(self, fig)
        # create a figure manager so that we can use pylab commands on the
        # main viewport
        def make_active(event):
            Gcf.set_active(self.manager)
        self.manager = FigureManagerQT(self, 1)
        self.manager._cidgcf = self.mpl_connect('button_press_event', make_active)
        Gcf.set_active(self.manager)
        self.setParent(parent)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.updateGeometry()
        # actually get key events
        self.setFocusPolicy(Qt.StrongFocus)
        self.mpl_connect('key_press_event', self.key_press)
        # These will not do anything in standalone mode, but do not hurt.
        if maincanvas:
            self.connect(session, SIGNAL('propsRequested'),
                         self.on_session_propsRequested)
            self.connect(session, SIGNAL('propsUpdated'),
                         self.on_session_propsUpdated)

    def on_session_propsRequested(self):
        session.props.canvas_logz = self.logz

    def on_session_propsUpdated(self):
        if 'canvas_logz' in session.props:
            self.logz = session.props.canvas_logz
            self.emit(SIGNAL('logzChanged'))

    def key_press(self, event):
        if key_press_handler:
            key_press_handler(event, self)

    def resizeEvent(self, event):
        # reimplemented to add tight_layout()
        w = event.size().width()
        h = event.size().height()
        dpival = float(self.figure.dpi)
        winch = w/dpival
        hinch = h/dpival
        self.figure.set_size_inches(winch, hinch)
        try:
            self.figure.tight_layout(pad=2)
        except Exception:
            pass
        self.plotter.save_layout()
        self.draw()
        self.update()
        QWidget.resizeEvent(self, event)

    def print_(self):
        sio = StringIO()
        self.print_figure(sio, format='svg')
        svg = QSvgRenderer(QByteArray(sio.getvalue()))
        sz = svg.defaultSize()
        aspect = sz.width()/float(sz.height())

        if self.printer is None:
            self.printer = QPrinter(QPrinter.HighResolution)
            self.printer.setOrientation(QPrinter.Landscape)

        dlg = QDialog(self)
        loadUi(dlg, 'printpreview.ui')
        dlg.width.setValue(self.print_width or 500)
        ppw = QPrintPreviewWidget(self.printer, dlg)
        dlg.layout().insertWidget(1, ppw)
        def render(printer):
            height = printer.height() * (dlg.width.value()/1000.)
            width = aspect * height
            painter = QPainter(printer)
            svg.render(painter, QRectF(0, 0, width, height))
        def sliderchanged(newval):
            ppw.updatePreview()
        self.connect(ppw, SIGNAL('paintRequested(QPrinter *)'), render)
        self.connect(dlg.width, SIGNAL('valueChanged(int)'), sliderchanged)
        if dlg.exec_() != QDialog.Accepted:
            return
        self.print_width = dlg.width.value()
        pdlg = QPrintDialog(self.printer, self)
        if pdlg.exec_() != QDialog.Accepted:
            return
        render(self.printer)


class MPLToolbar(NavigationToolbar2QT):

    icon_name_map = {
        'home.png':         'magnifier-zoom-fit.png',
        'back.png':         'arrow-180.png',
        'forward.png':      'arrow.png',
        'move.png':         'arrow-move.png',
        'zoom_to_rect.png': 'selection-resize.png',
        'filesave.png':     'document-pdf.png',
        'printer.png':      'printer.png',
        'pyconsole.png':    'terminal--arrow.png',
        'log-x.png':        'log-x.png',
        'log-y.png':        'log-y.png',
        'log-z.png':        'log-z.png',
        'exwindow.png':    'chart--arrow.png',
    }

    toolitems = list(NavigationToolbar2QT.toolitems)
    del toolitems[7]  # subplot adjust
    toolitems.insert(0, ('Log x', 'Logarithmic X scale', 'log-x', 'logx_callback'))
    toolitems.insert(1, ('Log y', 'Logarithmic Y scale', 'log-y', 'logy_callback'))
    toolitems.insert(2, ('Log z', 'Logarithmic Z scale for images', 'log-z',
                         'logz_callback'))
    toolitems.insert(3, (None, None, None, None))
    toolitems.append(('Print', 'Print the figure', 'printer',
                      'print_callback'))
    toolitems.append(('Pop out', 'Show the figure in a separate window',
                      'exwindow', 'popout_callback'))
    toolitems.append(('Execute', 'Show Python console', 'pyconsole',
                      'exec_callback'))

    def _init_toolbar(self):
        NavigationToolbar2QT._init_toolbar(self)
        self.locLabel.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
        self._actions['logx_callback'].setCheckable(True)
        self._actions['logy_callback'].setCheckable(True)
        self._actions['logz_callback'].setCheckable(True)
        self.connect(self.canvas, SIGNAL('logzChanged'),
                     self.on_canvas_logzChanged)

    def _icon(self, name):
        if name in self.icon_name_map:
            return QIcon(':/' + self.icon_name_map[name])
        return QIcon()

    def home(self):
        # always unzoom completely
        self._views.clear()
        self._positions.clear()
        self.canvas.figure.gca().autoscale()
        self.canvas.draw()
        return NavigationToolbar2QT.home(self)

    def logx_callback(self):
        ax = self.canvas.figure.gca()
        if ax.get_xscale() == 'linear':
            ax.set_xscale('symlog')
            self._actions['logx_callback'].setChecked(True)
        else:
            ax.set_xscale('linear')
            self._actions['logx_callback'].setChecked(False)
        self.canvas.draw()

    def logy_callback(self):
        ax = self.canvas.figure.gca()
        if ax.get_yscale() == 'linear':
            ax.set_yscale('symlog')
            self._actions['logy_callback'].setChecked(True)
        else:
            ax.set_yscale('linear')
            self._actions['logy_callback'].setChecked(False)
        self.canvas.draw()

    def logz_callback(self):
        ax = self.canvas.figure.gca()
        self.canvas.logz = not self.canvas.logz
        session.set_dirty()
        self._actions['logz_callback'].setChecked(self.canvas.logz)
        for im in ax.get_images():
            if self.canvas.logz:
                im.set_norm(LogNorm())
            else:
                im.set_norm(None)
        self.canvas.draw()

    def on_canvas_logzChanged(self):
        self._actions['logz_callback'].setChecked(self.canvas.logz)

    def print_callback(self):
        self.canvas.print_()

    def popout_callback(self):
        self.emit(SIGNAL('popoutRequested'))

    def exec_callback(self):
        try:
            from ufit.gui.console import ConsoleWindow
        except ImportError:
            logger.exception('Qt console window cannot be opened without '
                             'IPython; import error was:')
            QMessageBox.information(self, 'ufit',
                'Please install IPython with qtconsole to activate this function.')
            return
        w = ConsoleWindow(self)
        w.ipython.executeCommand('from ufit.lab import *')
        w.ipython.pushVariables({
            'fig': self.canvas.figure,
            'ax': self.canvas.figure.gca(),
            'D': [item for group in session.groups for item in group.items],
        })
        w.show()

    def save_figure(self, *args):
        filetypes = self.canvas.get_supported_filetypes_grouped()
        sorted_filetypes = sorted(filetypes.items())

        start = self.canvas.get_default_filename()
        filters = []
        for name, exts in sorted_filetypes:
            if 'eps' in exts or 'emf' in exts or 'jpg' in exts or \
                'pgf' in exts or 'raw' in exts:
                continue
            exts_list = " ".join(['*.%s' % ext for ext in exts])
            filter = '%s (%s)' % (name, exts_list)
            filters.append(filter)
        filters = ';;'.join(filters)
        fname = QFileDialog.getSaveFileName(self, 'Choose a filename to save to',
                                            start, filters)
        if fname:
            try:
                self.canvas.print_figure(unicode(fname))
            except Exception as e:
                logger.exception('Error saving file')
                QMessageBox.critical(self, 'Error saving file', str(e))


class SmallLineEdit(QLineEdit):
    def sizeHint(self):
        sz = QLineEdit.sizeHint(self)
        return QSize(sz.width()/1.5, sz.height())


class SettingGroup(object):
    def __init__(self, name):
        self.name = name
        self.settings = QSettings()

    def __enter__(self):
        self.settings.beginGroup(self.name)
        return self.settings

    def __exit__(self, *args):
        self.settings.endGroup()
        self.settings.sync()



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Backend using lmfit."""

from __future__ import absolute_import

from ufit.param import prepare_params, update_params

from lmfit import Parameters, minimize

__all__ = ['do_fit', 'backend_name']

backend_name = 'lmfit'

def do_fit(data, fcn, params, add_kw):

    # lmfit can handle expression-based parameters itself, but that is
    # a) buggy (cannot pass custom items into namespace without subclass)
    # and b) it is better to use the same mechanism in all backends

    x, y, dy = data.fit_columns
    meta = data.meta
    varying, varynames, dependent, _ = prepare_params(params, meta)

    lmfparams = Parameters()
    for p in varying:
        lmfparams.add(p.name, p.value, min=p.pmin, max=p.pmax)

    def lmfitfcn(lmfparams, data):
        pd = dict((pn, lmfparams[pn].value) for pn in varynames)
        update_params(dependent, meta, pd)
        return (fcn(pd, x) - y) / dy

    try:
        out = minimize(lmfitfcn, lmfparams, args=(data,), **add_kw)
    except Exception, e:
        return False, str(e), 0

    pd = dict((pn, lmfparams[pn].value) for pn in varynames)
    update_params(dependent, meta, pd)
    for p in params:
        p.value = pd[p.name]
        if p.name in lmfparams:
            p.error = lmfparams[p.name].stderr
            p.correl = lmfparams[p.name].correl
        else:
            p.error = 0
            p.correl = {}

    # sadly, out.message is a bit buggy
    message = ''
    if not out.success:
        message = out.lmdif_message
    if not out.errorbars:
        message += ' Could not estimate error bars.'
    return out.success, message, out.redchi



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Backend modules for different fitting packages/algorithms."""

from ufit import debug, UFitError

__all__ = ['set_backend', 'backend']

backend = None
available = []

try:
    from ufit.backends import scipy
except ImportError:
    scipy = None
else:
    backend = scipy
    available.append(scipy)

try:
    from ufit.backends import minuit
except ImportError:
    minuit = None
else:
    backend = minuit
    available.append(minuit)

try:
    from ufit.backends import lmfit
except ImportError:
    lmfit = None
else:
    backend = lmfit
    available.append(lmfit)

try:
    from ufit.backends import unifit
except ImportError:
    raise
    unifit = None
else:
    available.append(unifit)


def set_backend(which):
    """Select a new backend for fitting."""
    global backend
    backend = globals()[which]
    if backend is None:
        raise UFitError('Backend %r is not available' % which)
    debug('ufit using %s backend' % backend.backend_name)



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""
Resolution calculation using the Popovici method.  Based on
neutrons.instruments.tas.tasres from the neutrons Python package, compiled by
Marc Janoschek.
"""

import os
import multiprocessing

from numpy import pi, radians, degrees, sin, cos, tan, arcsin, arccos, \
     arctan2, abs, sqrt, real, matrix, diag, cross, dot, array, arange, \
     zeros, concatenate, reshape, delete, loadtxt
from numpy.random import randn
from numpy.linalg import inv, det, eig, norm


class unitcell(object):
    """
    Class which models a crystallographic unit cell from given lattice
    parameters and angles.  Further it provides functions to make basic
    calculations with the lattice vectors in the crystallographic and in the
    standard cartesian coordinate system.
    """

    def __init__(self, a, b, c, alpha, beta, gamma):
        """
        Instantiates the class object: necessary input parameters are the
        lattice constants a,b,c in Angstrom and the angles between the
        crystallographic axes, namely alpha, beta, gamma in degrees.
        """

        # unit cell parameters
        self.a = float(a)
        self.b = float(b)
        self.c = float(c)
        self.alpha = radians(alpha)
        self.beta  = radians(beta)
        self.gamma = radians(gamma)

        # often used values
        cosa = cos(self.alpha)
        cosb = cos(self.beta)
        cosg = cos(self.gamma)
        sing = sin(self.gamma)

        # conversion matrices
        # (s. International Tables of Crystallography, Volume B, 2nd Edition, p. 360)

        # phi is a needed constant
        phi = sqrt(1 - cosa**2 - cosb**2 - cosg**2 + 2*cosa*cosb*cosg)

        # from crystallographic (x) to cartesian setting (X) with first component
        # parallel to a and third to c* (reciprocal axis)
        # X = Mx, (x in fractional units=>dimensionless, X in dimension of length)
        self.crys2cartMat = matrix(
            [[a, b * cosg, c * cosb],
             [0, b * sing, c * (cosa - cosb*cosg)/sing],
             [0, 0,        c * phi/sing]]
        )

        # from cartesian (X) to crystallographic setting (x) with first
        # component parallel to a and third to c* (reciprocal axis)
        # x = M^(-1)X, (x in fractional units=>dimensionless, X in dimension of length)
        self.cart2crysMat = matrix(
            [[1./a, -1./(a*tan(self.gamma)), (cosa*cosg-cosb)/(a*phi*sing)],
             [0,    1/(b*sing),              (cosb*cosg-cosa)/(b*phi*sing)],
             [0,    0,                       sing/(c*phi)]]
        )

        #----- Real space lattice vectors

        self.a_vec = self.crys2cart(1, 0, 0).transpose()
        self.b_vec = self.crys2cart(0, 1, 0).transpose()
        self.c_vec = self.crys2cart(0, 0, 1).transpose()

        # this is needed if you only want to roatate from the cartesion to the
        # crystal system without changing the length
        self.crys2cartUnit = concatenate((self.a_vec/norm(self.a_vec),
                                          self.b_vec/norm(self.b_vec),
                                          self.c_vec/norm(self.c_vec)))
        self.cart2crysUnit = self.crys2cartUnit.I

        # Volume of unit cell
        self.V = dot(array(self.a_vec)[0],
                     cross(array(self.b_vec)[0], array(self.c_vec)[0]))

        #----- Reciprocal space lattice basis vectors
        self.a_star_vec = 2*pi*cross(array(self.b_vec)[0], array(self.c_vec)[0])/self.V
        self.b_star_vec = 2*pi*cross(array(self.c_vec)[0], array(self.a_vec)[0])/self.V
        self.c_star_vec = 2*pi*cross(array(self.a_vec)[0], array(self.b_vec)[0])/self.V

        #----- Reciprocal lattice index (r.l.u.) to cartesian matrix
        self.rlu2cartMat = concatenate((matrix(self.a_star_vec).T,
                                        matrix(self.b_star_vec).T,
                                        matrix(self.c_star_vec).T), 1)
        self.cart2rluMat = inv(self.rlu2cartMat)

        #----- Reciprocal lattice constants; note that there is a factor
        # 2*pi difference to CrysFML definition
        self.a_star = sqrt(dot(self.a_star_vec, self.a_star_vec))
        self.b_star = sqrt(dot(self.b_star_vec, self.b_star_vec))
        self.c_star = sqrt(dot(self.c_star_vec, self.c_star_vec))

        #----- Reciprocal lattice angles
        self.ralpha = arccos(dot(self.b_star_vec, self.c_star_vec)/(self.b_star*self.c_star))
        self.rbeta = arccos(dot(self.a_star_vec, self.c_star_vec)/(self.a_star*self.c_star))
        self.rgamma = arccos(dot(self.b_star_vec, self.a_star_vec)/(self.b_star*self.a_star))

        #----- Reciprocal cell volume
        self.VR = dot(self.a_star_vec, cross(self.b_star_vec, self.c_star_vec))

        #----- Review this array, I think it has to be transposed to be useful
        self.Q2c = matrix([self.a_star_vec, self.b_star_vec, self.c_star_vec])

        #----- Metric tensors
        self.GD = zeros([3,3])
        self.GR = zeros([3,3])

        # real space
        self.GD[0,0] = self.a**2
        self.GD[1,1] = self.b**2
        self.GD[2,2] = self.c**2

        self.GD[0,1] = self.a * self.b * cos(self.gamma)
        self.GD[0,2] = self.a * self.c * cos(self.beta)
        self.GD[1,2] = self.b * self.c * cos(self.alpha)

        self.GD[1,0] = self.GD[0,1]
        self.GD[2,0] = self.GD[0,2]
        self.GD[2,1] = self.GD[1,2]

        self.GD = matrix(self.GD)

        # reciprocal space
        self.GR[0,0] = self.a_star**2
        self.GR[1,1] = self.b_star**2
        self.GR[2,2] = self.c_star**2

        self.GR[0,1] = self.a_star * self.b_star * cos(self.rgamma)
        self.GR[0,2] = self.a_star * self.c_star * cos(self.rbeta)
        self.GR[1,2] = self.b_star * self.c_star * cos(self.ralpha)

        self.GR[1,0] = self.GR[0,1]
        self.GR[2,0] = self.GR[0,2]
        self.GR[2,1] = self.GR[1,2]

        self.GR = matrix(self.GR)

    def __str__(self):
        """Generates string for output on console or file with information about
        unitcell.
        """
        repstr = "<Unit cell object>\n"
        repstr += "a      b      c      alpha   beta   gamma\n"
        degree_sym = "deg"
        repstr += "%2.3fA %2.3fA %2.3fA %3.1f%s   %3.1f%s   %3.1f%s\n" % \
            (self.a, self.b, self.c, degrees(self.alpha), degree_sym,
             degrees(self.beta), degree_sym, degrees(self.gamma), degree_sym)
        repstr += "a*     b*     c*     alpha*  beta*  gamma*\n"
        repstr += "%2.3fA-1 %2.3fA-1 %2.3fA-1 %3.1f%s   %3.1f%s   %3.1f%s\n" % \
            (self.a_star, self.b_star, self.c_star, degrees(self.ralpha), degree_sym,
             degrees(self.rbeta), degree_sym, degrees(self.rgamma), degree_sym)
        return repstr

    __repr__ = __str__

    def __eq__(self, other):
        if not isinstance(other, unitcell):
            return False
        return self.a == other.a and \
            self.b == other.b and \
            self.c == other.c and \
            self.alpha == other.alpha and \
            self.beta == other.beta and \
            self.gamma == other.gamma

    def crys2cart(self, rx, ry, rz):
        """Calculates vectors in the cartesian coordinate frame in units of
        Angstrom from given (rx,ry,rz) in fractional units in the
        crystallographic frame.
        """
        x = matrix([[float(rx)],[float(ry)],[float(rz)]])
        return self.crys2cartMat * x

    def cart2crys(self, rx, ry, rz):
        """Calculates vectors in the crystallographic frame in fractional units
        from given (rx,ry,rz) in the cartesian frame in units of Angstrom.
        """
        X = matrix([[float(rx)], [float(ry)], [float(rz)]])
        return self.cart2crysMat * X

    def QCart(self, h, k, l):
        """Calculates Q vector in reciprocal space lattice vector."""
        return h*self.a_star_vec + k*self.b_star_vec + l*self.c_star_vec

    def QCartMag(self, h, k, l):
        """Calculates magnitude of the Q vector in units of Angstrom if one
        provides it in hkl [r.l.u.].
        """
        return norm(self.QCart(h, k, l))

    def calctheta(self, h, k, l, lam):
        """Calculates theta angle (in degrees) of Bragg reflection from lattice
        constants and given h, k, l and lambda (in Angstrom).
        """
        Q = self.QCartMag(h,k,l)
        return degrees(arcsin(Q*lam/4/pi))


# parameters
PARNAMES = [
    'dm',
    'da',
    'etam',
    'etas',
    'etaa',
    'sm',
    'ss',
    'sa',
    'k',
    'kfix',
    'alpha1',
    'alpha2',
    'alpha3',
    'alpha4',
    'beta1',
    'beta2',
    'beta3',
    'beta4',
    'as',
    'bs',
    'cs',
    'aa',
    'bb',
    'cc',
    'ax',
    'ay',
    'az',
    'bx',
    'by',
    'bz',
    # 'qx',    # these are irrelevant for our resolution calc.
    # 'qy',
    # 'qz',
    # 'en',
    # 'dqx',
    # 'dqy',
    # 'dqz',
    # 'de',
    # 'gh',
    # 'gk',
    # 'gl',
    # 'gmod',
]

CFGNAMES = [
    'sourcetype',
    'sourcewidth',
    'sourceheight',
    'guideon',
    'guidedivh',
    'guidedivv',
    'samplecuboid',
    'samplewidthperp',
    'samplewidthpar',
    'sampleheight',
    'detrectangular',
    'detwidth',
    'detheight',
    'monothickness',
    'monowidth',
    'monoheight',
    'anathickness',
    'anawidth',
    'anaheight',
    'distsm',
    'distms',
    'distsa',
    'distad',
    'curvmonoh',
    'curvmonov',
    'curvanah',
    'curvanav',
    'distmoni',
    'moniwidth',
    'moniheight',
]


# instrumental setup description
CFGDESC = [
    "=0 for circular source, =1 for rectangular source",
    "width/diameter of the source (cm)",
    "height/diameter of the source (cm)",
    "=0 no Guide, =1 for Guide",
    "horizontal guide divergence (minutes/Angs)",
    "vertical guide divergence (minutes/Angs)",
    "=0 for cylindrical sample, =1 for cuboid sample",
    "sample width/diameter perp. to Q (cm)",
    "sample width/diameter along Q (cm)",
    "sample height (cm)",
    "=0 for circular detector, =1 for rectangular detector",
    "width/diameter of the detector (cm)",
    "height/diameter of the detector (cm)",
    "thickness of monochromator (cm)",
    "width of monochromator (cm)",
    "height of monochromator (cm)",
    "thickness of analyser (cm)",
    "width of analyser (cm)",
    "height of analyser (cm)",
    "distance between source and monochromator (cm)",
    "distance between monochromator and sample (cm)",
    "distance between sample and analyser (cm)",
    "distance between analyser and detector (cm)",
    "horizontal curvature of monochromator 1/radius (cm-1)",
    "vertical curvature of monochromator (cm-1)",
    "horizontal curvature of analyser (cm-1)",
    "vertical curvature of analyser (cm-1)",
    "distance monochromator-monitor",
    "width monitor",
    "height monitor"
]

MEV2AA2 = 0.48259642  # hbar**2/(2 * m_n)
MIN2RAD = 1/60. * pi/180.


class resmat(object):
    """Class which calculates the resolution matrix, and the renormalisation
    volume R0 for a given triple axis parameter set which describes a specific
    setup at a specific instrument by using the Popovici method (REF).

    Based on these two properties several others can be calculated.

    This code was ported from Bertrand Roessli's octave version of the ILL
    rescal MATLAB tools.
    Checked against rescal5 rc_popma.m from ILL matlab libraries from 06/2013.
    """

    def __init__(self, cfg, par):
        # instrument definition (lengths etc. defining a specific instrument)
        self.cfg = list(cfg)
        # experimental setup (plus easily changed instrument things)
        self.par = par.copy()

        # use a fixed resolution matrix
        self.fixed_res = False

        # a string message or None if calculations are fine
        self.ERROR = None

        # magnitude of scattering vector
        #----- calculate q0
        #-----input a, b, c; alpha, beta, gamma; qx, qy, qz
        self.unitc = unitcell(self.par['as'], self.par['bs'], self.par['cs'],
                              self.par['aa'], self.par['bb'], self.par['cc'])
        self.q0 = self.unitc.QCartMag(self.par['qx'], self.par['qy'], self.par['qz'])
        self.Q2c = self.unitc.Q2c

        # resolution
        self.R0 = 0
        self.N = matrix(zeros((4,4)))
        # calculate resolution matrix
        self.calc_popovici()
        self.calc_STrafo()

        self._cache = {}

    def calc_popovici(self):
        """Performs the actual calculation.

        """
        # reset errors before new calculation
        self.ERROR = None

        #----- recalculate q0
        #-----input a, b, c; alpha, beta, gamma; qx, qy, qz
        self.q0 = self.unitc.QCartMag(self.par['qx'], self.par['qy'], self.par['qz'])

        #----- INPUT SPECTROMETER PARAMETERS.
        q0 = self.q0

        dm = self.par['dm']               # monochromator d-spacing in Angs
        da = self.par['da']               # analyser d-spacing in Angs
        etam = self.par['etam']*MIN2RAD   # monochromator mosaic
        etamp = etam                      # vertical mosaic spread of the monochromator
        etaa = self.par['etaa']*MIN2RAD   # analyser mosaic
        etaap = etaa                      # vertical mosaic spread of the analyser
        etas = self.par['etas']*MIN2RAD   # sample mosaic
        etasp = etas                      # vertical mosaic spread of the sample
        sm = self.par['sm']               # scattering sense of monochromator [left = +1,right = -1]
        ss = self.par['ss']               # scattering sense of sample [left = +1,right = -1]
        sa = self.par['sa']               # scattering sense of analyser [left = +1,right = -1]
        kfix = self.par['k']              # fixed momentum component in ang-1
        fx = self.par['kfix']             # fx = 1 for fixed incident and 2 for scattered wavevector
        # collimation: all input angles are in minutes of arc
        alf0 = self.par['alpha1']*MIN2RAD # horizontal pre-monochromator
        alf1 = self.par['alpha2']*MIN2RAD # horizontal pre-sample
        alf2 = self.par['alpha3']*MIN2RAD # horizontal post-sample
        alf3 = self.par['alpha4']*MIN2RAD # horizontal post-analyser
        bet0 = self.par['beta1']*MIN2RAD  # vertical pre-monochromator
        bet1 = self.par['beta2']*MIN2RAD  # vertical pre-sample
        bet2 = self.par['beta3']*MIN2RAD  # vertical post-sample
        bet3 = self.par['beta4']*MIN2RAD  # vertical post-analyser
        w = self.par['en']                # energy transfer in meV

        # _____________________Instrument Parameters________________________________________

        nsou = self.cfg[0]                # =0 for circular source, =1 for rectangular source.
        ysrc = self.cfg[1]                # width/diameter of the source [cm].
        zsrc = self.cfg[2]                # height/diameter of the source [cm].

        flag_guide = self.cfg[3]          # =0 for no guide, =1 for guide.
        guide_h = self.cfg[4]             # horizontal guide divergence [mins/Angs]
        guide_v = self.cfg[5]             # vertical guide divergence [mins/Angs]

        nsam = self.cfg[6]                # =0 for cylindrical sample, =1 for cuboid sample.
        xsam = self.cfg[7]                # sample width/diameter perp. to Q [cm].
        ysam = self.cfg[8]                # sample width/diameter along Q [cm].
        zsam = self.cfg[9]                # sample height [cm].

        ndet = self.cfg[10]               # =0 for circular detector, =1 for rectangular detector.
        ydet = self.cfg[11]               # width/diameter of the detector [cm].
        zdet = self.cfg[12]               # height/diameter of the detector [cm].

        xmon = self.cfg[13]               # thickness of monochromator [cm].
        ymon = self.cfg[14]               # width of monochromator [cm].
        zmon = self.cfg[15]               # height of monochromator [cm].

        xana = self.cfg[16]               # thickness of analyser [cm].
        yana = self.cfg[17]               # width of analyser [cm].
        zana = self.cfg[18]               # height of analyser [cm].

        L0 = self.cfg[19]                 # distance between source and monochromator [cm].
        L1 = self.cfg[20]                 # distance between monochromator and sample [cm].
        L2 = self.cfg[21]                 # distance between sample and analyser [cm].
        L3 = self.cfg[22]                 # distance between analyser and detector [cm].

        # curvature radii are handled below, after we calculated thetam and thetaa.

        L1mon = self.cfg[27]              #distance monochromator monitor [cm]
        monitorw = self.cfg[28]/sqrt(12)  #monitor width [cm]
        monitorh = self.cfg[29]/sqrt(12)  #monitor height [cm]

        f16 = 1/16.
        f12 = 1/12.

        # _____________________________________________________________________________

        # THIS ASSUMES w IS IN meV!

        # Calculate ki and kf, thetam and thetaa
        ki = abs(sqrt(kfix**2 + (fx-1) * MEV2AA2 * w))  # kinematical equations.
        kf = abs(sqrt(kfix**2 - (2-fx) * MEV2AA2 * w))

        # Test if scattering triangle is closed

        cos_2theta = (ki**2 + kf**2 - q0**2) / (2*ki*kf)
        # print cos_2theta
        if not -1 <= cos_2theta <= 1:
            self.ERROR = 'scattering triangle not closed'
            return

        thetaa = sa*arcsin(pi/(da*kf))      # theta angles for analyser
        thetam = sm*arcsin(pi/(dm*ki))      # and monochromator.
        thetas = ss*0.5*arccos((ki**2 + kf**2 - q0**2) / (2*ki*kf)) # scattering angle from sample.
        phi = arctan2(-kf*sin(2*thetas), ki - kf*cos(2*thetas))

        # _____________________________________________________________________________

        # automatic determination of curvature, added GB 06/2013 (not in RESCAL5)
        if self.cfg[23] == -1:
            if flag_guide:
                romh = sm*0.5*(1./L1)*sin(abs(thetam))
            else:
                romh = sm*0.5*(1./L0 + 1./L1)*sin(abs(thetam))
        else:
            romh = sm*self.cfg[23]          # horizontal curvature of monochromator 1/radius [cm-1].
        if self.cfg[24] == -1:
            if flag_guide:
                romv = sm*0.5*(1./L1)*sin(abs(thetam))**(-1)
            else:
                romv = sm*0.5*(1./L0 + 1./L1)*sin(abs(thetam))**(-1)
        else:
            romv = sm*self.cfg[24]          # vertical curvature of monochromator [cm-1].
        if self.cfg[25] == -1:
            roah = sa*0.5*(1./L2 + 1./L3)*sin(abs(thetaa))
        else:
            roah = sa*self.cfg[25]          # horizontal curvature of analyser [cm-1].
        if self.cfg[26] == -1:
            roav = sa*0.5*(1./L2 + 1./L3)*sin(abs(thetaa))**(-1)
        else:
            roav = sa*self.cfg[26]          # vertical curvature of analyser [cm-1].

        # _____________________________________________________________________________

        # Make up the matrices in appendix 1 of M. Popovici (1975).

        # First check for incident guide

        if flag_guide == 1:
            alf0_guide = MIN2RAD*2*pi*guide_h/ki
            bet0 = MIN2RAD*2*pi*guide_v/ki

            if alf0_guide <= alf0:  # take into account collimator in the guide
                alf0 = alf0_guide   # (not in RESCAL5)

        G = matrix(zeros((8, 8)))
        G[0,0] = 1/alf0**2    # horizontal and vertical collimation matrix. The 4 Soller collimators
        G[1,1] = 1/alf1**2    # are described by a 8x8 diagonal matrix with non-zero elements.
        G[2,2] = 1/bet0**2    # alfi and beti are the FWHM horizontal and vertical beam divergences in
        G[3,3] = 1/bet1**2    # the collimators
        G[4,4] = 1/alf2**2
        G[5,5] = 1/alf3**2
        G[6,6] = 1/bet2**2
        G[7,7] = 1/bet3**2

        F = matrix(zeros((4, 4)))
        F[0,0] = 1/etam**2    # monochromator and analyser mosaic matrix. horizontal and vertical mosaic
        F[1,1] = 1/etamp**2   # spreads for monochromator and analyzer crystals
        F[2,2] = 1/etaa**2
        F[3,3] = 1/etaap**2

        C = matrix(zeros((4, 8)))
        C[0,0] = 1./2
        C[0,1] = C[0,0]
        C[2,4] = C[0,0]
        C[2,5] = C[0,0]
        C[1,2] = 1/(2*sin(thetam))    # thetam = scattering angle of monochromator
        C[1,3] = -C[1,2]              # thetam = -arcsin(Tm/2ki))epsilonm;
                                      # Tm(Ta) = 2Pi/dm(da) mono and ana scattering vectors
        C[3,6] = 1/(2*sin(thetaa))    # thetaa = scattering angle of analyzer
        C[3,7] = -C[3,6]              # thetaa = -arcsin(Ta/2kf))epsilona
                                      # epsilonm/a = 1 if sample scattering
                                      # direction opposite do mono/ana scattering dir, -1 otherwise
        A = matrix(zeros((6, 8)))
        A[0,0] = ki/(tan(thetam)*2)
        A[0,1] = -A[0,0]
        A[1,1] = ki
        A[2,3] = A[1,1]
        A[3,4] = kf/(tan(thetaa)*2)
        A[3,5] = -A[3,4]
        A[4,4] = kf
        A[5,6] = A[4,4]

        B = matrix(zeros((4, 6)))
        B[0,0] = cos(phi)
        B[0,1] = sin(phi)
        B[0,3] = -cos(phi-2*thetas)    # thetas = scattering angle of
        B[0,4] = -sin(phi-2*thetas)    # 2thetas = arccos[(ki2 + kf2 - Q2)/(2kikf)]
        B[1,0] = -B[0,1]
        B[1,1] = B[0,0]
        B[1,3] = -B[0,4]
        B[1,4] = B[0,3]
        B[2,2] = 1.
        B[2,5] = -1.
        B[3,0] = 2*ki/MEV2AA2
        B[3,3] = -2*kf/MEV2AA2

        # Now include the spatial effects.

        # --- monochromator spatial covariances ----------
        S1I = matrix(zeros((3, 3)))
        factor = f12
        S1I[0,0] = factor*xmon**2
        S1I[1,1] = factor*ymon**2
        S1I[2,2] = factor*zmon**2

        # --- sample spatial covariances -----------
        S2I = zeros((3, 3))
        if nsam == 0:
            factor = f16
        else:
            factor = f12

        S2I[0,0] = factor*xsam**2
        S2I[1,1] = factor*ysam**2
        S2I[2,2] = f12*zsam**2

        # --- analyser spatial covariances ---------
        factor = f12
        S3I = zeros((3, 3))
        S3I[0,0] = factor*xana**2
        S3I[1,1] = factor*yana**2
        S3I[2,2] = factor*zana**2
        # --- construct the full spatial covariance matrix
        SI = zeros((13, 13))
        # --- source covariance --------------------
        if nsou == 0:
            factor = f16      # factor for converting diameter**2 to variance**2
        else:
            factor = f12      # factor for converting width**2 to variance**2

        SI[0,0] = factor*ysrc**2
        SI[1,1] = factor*zsrc**2
        # --- add in the mono. sample & analyser ---
        SI[2:5][:,2:5] = S1I
        SI[5:8][:,5:8] = S2I
        SI[8:11][:,8:11] = S3I
        # --- detector covariance ------------------
        if ndet == 0:
            factor = f16
        else:
            factor = f12

        SI[11,11] = factor*ydet**2
        SI[12,12] = factor*zdet**2

        S = inv(5.545 * SI)

        T = matrix(zeros((4, 13)))
        T[0,0] = -1./(2*L0)
        T[0,2] = cos(thetam)*(1./L1 - 1./L0)/2
        T[0,3] = sin(thetam)*(1./L0 + 1./L1 - 2*romh/sin(thetam))/2
        T[0,5] = sin(thetas)/(2.*L1)
        T[0,6] = cos(thetas)/(2.*L1)
        T[1,1] = -1./(2*L0*sin(thetam))
        T[1,4] = (1./L0 + 1./L1 - 2*sin(thetam)*romv)/(2*sin(thetam))
        T[1,7] = -1./(2*L1*sin(thetam))
        T[2,5] = sin(thetas)/(2.*L2)
        T[2,6] = -cos(thetas)/(2.*L2)
        T[2,8] = cos(thetaa)*(1./L3 - 1./L2)/2
        T[2,9] = sin(thetaa)*(1./L2 + 1./L3 - 2*roah/sin(thetaa))/2
        T[2,11] = 1./(2*L3)
        T[3,7] = -1./(2*L2*sin(thetaa))
        T[3,10] = (1./L2 + 1./L3 - 2*sin(thetaa)*roav)/(2*sin(thetaa))
        T[3,12] = -1./(2*L3*sin(thetaa))

        D = matrix(zeros((8, 13)))
        D[0,0] = -1./L0
        D[0,2] = -cos(thetam)/L0
        D[0,3] = sin(thetam)/L0
        D[2,1] = D[0,0]
        D[2,4] = -D[0,0]
        D[1,2] = cos(thetam)/L1
        D[1,3] = sin(thetam)/L1
        D[1,5] = sin(thetas)/L1
        D[1,6] = cos(thetas)/L1
        D[3,4] = -1./L1
        D[3,7] = -D[3,4]
        D[4,5] = sin(thetas)/L2
        D[4,6] = -cos(thetas)/L2
        D[4,8] = -cos(thetaa)/L2
        D[4,9] = sin(thetaa)/L2
        D[6,7] = -1./L2
        D[6,10] = -D[6,7]
        D[5,8] = cos(thetaa)/L3
        D[5,9] = sin(thetaa)/L3
        D[5,11] = 1./L3
        D[7,10] = -D[5,11]
        D[7,12] = D[5,11]

        # Construction of the resolution matrix
        #if det(D*inv(S+T.transpose()*F*T)*D.transpose()) != 0:
        MI = B*A*inv(inv(D*inv(S + T.T*F*T)*D.T) + G)*A.T*B.T # including spatial effects.
        #else:
        #MI = B*A*(inv(G))*A.transpose()*B.transpose()
        #MI = B*A*[inv[G+C'*F*C]]*A'*B'  # Cooper and Nathans matrix,
        MI[1,1] = MI[1,1] + q0**2*etas**2
        MI[2,2] = MI[2,2] + q0**2*etasp**2
        M = inv(MI)
        NP = 5.545*M                        # Correction factor as input parameters
        N = NP

        #----- Normalisation factor

        #mon = 1          # monochromator reflectivity
        #ana = 1          # detector and analyser crystal efficiency function. [const.]

        #if mon_flag==1
        #   vi = 1
        #else
        #   vi = mon*ki**3*cot[thetam]*15.75*bet0*bet1*etam*alf0*alf1
        #   vi = vi/sqrt[[2*sin[thetam]*etam]**2+bet0**2+bet1**2]
        #   vi = abs[vi/sqrt[alf0**2+alf1**2+4*etam**2]]
        #end
        #
        #vf = ana*kf**3*cot[thetaa]*15.75*bet2*bet3*etaa*alf2*alf3
        #vf = vf/sqrt[[2*sin[thetaa]*etaa]**2+bet2**2+bet3**2]
        #vf = abs[vf/sqrt[alf2**2+alf3**2+4*etaa**2]]
        #
        #R0 = vi*vf*sqrt[det[NP]]/[2*pi]**2  # Dorner form of resolution normalisation
        #                                  # see Mitchell, Cowley and Higgins.

        #Taken from Zheludev's version [RESLIB 3.1]
        #Calculation of prefactor, normalized to source

        Rm = ki**3/tan(thetam)
        Ra = kf**3/tan(thetaa)
        R0 = Rm*Ra*(2*pi)**4/(64*pi**2*sin(thetam)*sin(thetaa))* \
            sqrt(det(F) / det(inv(D*inv(S + T.T*F*T)*D.T) + G)) #Popovici
        # Werner and Pynn correction
        # for mosaic spread of crystal.
        R0 = abs(R0 / (etas*sqrt(1/etas**2 + q0**2*N[1,1])))

        RM_ = matrix(zeros((4, 4)))
        RM_[0,0] = M[0,0]
        RM_[1,0] = M[1,0]
        RM_[0,1] = M[0,1]
        RM_[1,1] = M[1,1]
        RM_[0,2] = M[0,3]
        RM_[2,0] = M[3,0]
        RM_[2,2] = M[3,3]
        RM_[2,1] = M[3,1]
        RM_[1,2] = M[1,3]
        RM_[0,3] = M[0,2]
        RM_[3,0] = M[2,0]
        RM_[3,3] = M[2,2]
        RM_[3,1] = M[2,1]
        RM_[1,3] = M[1,2]

        #Transform prefactor to Chesser-Axe normalization
        R0 = R0/(2*pi)**2*sqrt(det(RM_))

        #---------------------------------------------------------------------------------------------
        #Include kf/ki part of cross section
        R0 = R0*kf/ki

        #include monitor efficiency
        #Normalisation to flux monitor
        #bshape = matrix(diag((ysrc,zsrc)))
        #mshape = matrix(diag((xmon,ymon,zmon)))
        #monitorshape = matrix(diag((monitorw,monitorh)))
        g = G[0:4][:,0:4]
        f = F[0:2][:,0:2]
        #c = C[0:2][:,0:4]
        t = matrix(zeros((2,7)))
        t[0,0] = -1./(2*L0)  #mistake in paper
        t[0,2] = cos(thetam)*(1./L1mon - 1./L0)/2
        t[0,3] = sin(thetam)*(1./L0 + 1./L1mon - 2*romh/(sin(thetam)))/2
        t[0,6] = 1./(2*L1mon)
        t[1,1] = -1./(2*L0*sin(thetam))
        t[1,4] = (1./L0 + 1./L1mon - 2*sin(thetam)*romv)/(2*sin(thetam))
        sinv = matrix(diag((ysrc, zsrc, xmon, ymon, zmon, monitorw, monitorh))) #S-1 matrix
        s = sinv**(-1)
        d = matrix(zeros((4, 7)))
        d[0,0] = -1./L0
        d[0,2] = -cos(thetam)/L0
        d[0,3] = sin(thetam)/L0
        d[2,1] = D[0,0]
        d[2,4] = -D[0,0]
        d[1,2] = cos(thetam)/L1mon
        d[1,3] = sin(thetam)/L1mon
        d[1,5] = 0
        d[1,6] = 1./L1mon
        d[3,4] = -1./L1mon
        Rmon = Rm*(2*pi)**2/(8*pi*sin(thetam))*sqrt(det(f)/det((d*(s + t.T*f*t)**(-1)*d.T)**(-1) + g))
        R0 = R0*ki #1/ki monitor efficiency
        R0 = R0/Rmon
        #------ END OF ZHELUDEV's CORRECTION

        #----- Final error check
        if NP.imag.all() == 0:
            self.ERROR = None
        else:
            self.ERROR = 'problem with matrix calculation'
            return

        self.R0 = abs(R0)
        self.NP = NP
        #return (R0, NP, vi, vf, Error)

    def __str__(self):
        """Returns info string about resolution matrix."""
        info = self.par.copy()
        if self.par['kfix'] == 1:
            info['efixstr'] = 'fixed incident energy k_i = %2.4f A-1' % self.par['k']
        else:
            info['efixstr'] = 'fixed final energy k_f = %2.4f A-1' % self.par['k']
        info['q0'] = self.q0
        if not self.ERROR:
            mat = self.NP.tolist()
            info['mat'] = '\n'.join(''.join(('%5.2f' % mat[i][j]).rjust(10)
                                            for j in range(4)) for i in range(4))
            # Calculate Bragg width in direction of scan and for w-scan, and
            # corresponding Lorentz factors
            bragw = self.calcBragg()
            info['brqx'], info['brqy'], info['brqz'], info['brva'], info['brde'] = bragw
            info['R0'] = self.R0

        p1 =  """\
Resolution matrix for a triple axis spectrometer calculated by the Popovici method:

Spectrometer Setup:
===================
d-spacings:   dm = %(dm)1.4f    da = %(da)1.4f
mosaic    : etam = %(etam)1.4f etas = %(etas)1.4f etaa = %(etaa)1.4f
s-sense   :   sm = %(sm)-4i      ss = %(ss)-4i      sa = %(sa)-4i
alpha 1->4:  %(alpha1)i-Mono-%(alpha2)i-Sample-%(alpha3)i-Ana-%(alpha4)i (horizontal collimation)
beta  1->4:  %(beta1)i-Mono-%(beta2)i-Sample-%(beta3)i-Ana-%(beta4)i   (vertical collimation)

Sample Parameters:
==================
Lattice information:
a         b         c         alpha      beta       gamma
%(as)2.3f A   %(bs)2.3f A   %(cs)2.3f A   %(aa)3.1f deg   %(bb)3.1f deg   %(cc)3.1f deg

Scattering plane:
AX      AY      AZ      BX      BY      BZ
%(ax)1.3f   %(ay)1.3f   %(az)1.3f   %(bx)1.3f   %(by)1.3f   %(bz)1.3f

%(efixstr)s
reciprocal space position: qh = %(qx)1.3f qk = %(qy)1.3f ql = %(qz)1.3f (r.l.u.) en = %(en)2.3f (meV)
scattering vector Q = %(q0)2.5f A-1

Resolution Info:
================
""" % info
        if self.ERROR:
            p2 = 'ERROR: ' + self.ERROR
        else:
            p2 = """\
=> Resolution Volume: R0 = %(R0)3.3f A-3*meV

=> Resolution Matrix (in frame Qx, Qy, Qz, E):
%(mat)s

=> Bragg widths:
   Qx       Qy       Qz (A-1)  Vanadium  dE (meV)
   %(brqx)1.5f  %(brqy)1.5f  %(brqz)1.5f   %(brva)1.5f   %(brde)1.5f
""" % info
        return p1 + p2

    __repr__ = __str__

    def run_mcstas(self, NMC, QE):
        os.system('rm -rf /tmp/x')
        cmd = '/data/software/ufit/mcstas/templateTAS.out -n%g' % (NMC * 1000)
        cmd += ' --dir /tmp/x'
        for p in self.par:
            if p in ('qx', 'qy', 'qz', 'en', 'de', 'dqx', 'dqy', 'dqz', 'gh',
                     'gk', 'gl', 'gmod', 'etas'):
                continue
            pn = p.upper()
            if p == 'k':
                pn = 'KFIX'
            if p == 'kfix':
                pn = 'FX'
            cmd += ' %s=%s' % (pn, self.par[p])
        cmd += ' L1=%s L2=%s L3=%s L4=%s' % (self.cfg[19]/100,
                                             self.cfg[20]/100,
                                             self.cfg[21]/100,
                                             self.cfg[22]/100)
        cmd += ' WM=%s HM=%s WA=%s HA=%s WD=%s HD=%s' % (self.cfg[14]/100,
                                                         self.cfg[15]/100,
                                                         self.cfg[17]/100,
                                                         self.cfg[18]/100,
                                                         self.cfg[11]/100,
                                                         self.cfg[12]/100)
        #if self.cfg[23] <= 0:
        #    cmd += ' RMH=%s' % self.cfg[23]
        #else:
        #    cmd += ' RMH=%s' % (1/self.cfg[23]/100)
        #if self.cfg[24] <= 0:
        #    cmd += ' RMV=%s' % self.cfg[24]
        #else:
        #    cmd += ' RMV=%s' % (1/self.cfg[24]/100)
        #if self.cfg[25] <= 0:
        #    cmd += ' RAH=%s' % self.cfg[25]
        #else:
        #    cmd += ' RAH=%s' % (1/self.cfg[25]/100)
        #if self.cfg[26] <= 0:
        #    cmd += ' RAV=%s' % self.cfg[26]
        #else:
        #    cmd += ' RAV=%s' % (1/self.cfg[26]/100)
        cmd += ' swidth=%s sheight=%s sthick=%s' % (self.cfg[7]/100,
                                                    self.cfg[8]/100,
                                                    self.cfg[9]/100)
        cmd += ' QH=%s QK=%s QL=%s EN=%s' % tuple(QE)
        print '[MCSTAS] running:', cmd
        os.system(cmd)
        print '[MCSTAS] finished'
        try:
            arr = loadtxt('/tmp/x/res.dat', ndmin=2)
        except IOError:
            return [], [], [], [], []
        # some mcstas bug?!
        if len(arr.T) == 4:
            return [], [], [], [], []
        kix, kiy, kiz, kfx, kfy, kfz, x, y, z, pi, pf = arr.T
        # NOTE: cyclic shift of x, y, z here due to different coordinate system
        # conventions in McStas and here:
        # - McStas: z along beam, y upwards, x completes RHS
        # - here:   x along beam, z upwards, y completes RHS
        Q = array((kiz, kix, kiy)) - array((kfz, kfx, kfy))
        w = 2.072*(kix**2 + kiy**2 + kiz**2 - kfx**2 - kfy**2 - kfz**2)
        p = pi*pf/1e7
        print '%d neutrons for MC calculation' % len(w)
        Q = array([array(dot(self.unitc.cart2rluMat, q))[0] for q in Q.T]).T
        return Q[0], Q[1], Q[2], w, p

    def calc_STrafo(self):
        """Calculates transformation matrix self.S, which transforms from the
        coordinate frame given in h, k, l (r.l.u.) and E(meV) to cartesian
        crystallographic axis in units of Angstrom^{-1} and meV (Matrix
        self.Q2c; s neutrons.crystal.unitcell modul).

        Then a second transformation into the coordinate frame defined by V1,V2
        and V3 is performed (Matrix U). V1 and V2 are the vector defining the
        scattering plane in the TAS experiments, V3 is perpendicular to the
        scattering plane (zone axis).
        """

        #----- Work out transformations
        A1 = matrix([self.par['ax'], self.par['ay'], self.par['az']]).transpose()
        A2 = matrix([self.par['bx'], self.par['by'], self.par['bz']]).transpose()
        V1 = self.Q2c*A1
        V2 = self.Q2c*A2
        #----- Form unit vectors V1, V2, V3 in scattering plane

        V3 = cross(V1.transpose(),V2.transpose())
        V2 = cross(V3,V1.transpose())
        V3 = V3/sqrt(dot(V3,V3.transpose()))
        V2 = V2/sqrt(dot(V2,V2.transpose()))
        V1 = V1.transpose()/sqrt(dot(V1.transpose(),V1))
        V1 = array(V1)[0]
        V2 = array(V2)[0]
        V3 = array(V3)[0].transpose()
        U = matrix([V1, V2, V3])
        self.S = U*self.Q2c

    def sethklen(self, h, k, l, en):
        self.par['qx'] = h
        self.par['qy'] = k
        self.par['qz'] = l
        self.par['en'] = en
        self.calc_popovici()
        self.calc_STrafo()

    def calcSigma(self):
        #calculates sigmas in system h,k,l (r.l.u) en (meV)
        #print self.M
        [E, V] = eig(self.M)

        #print 'Eigenvalues'
        #print E
        #print 'eigenvecs'
        #print V

        E = matrix(real(diag(E)))
        sigma = zeros((1, 4))[0]
        sigma[0] = real(1./sqrt(E[0,0]))
        sigma[1] = real(1./sqrt(E[1,1]))
        sigma[2] = real(1./sqrt(E[2,2]))
        sigma[3] = real(1./sqrt(E[3,3]))
        self.b_mat = reshape(inv(V.transpose().transpose()), (1, 16))
        self.sigma = sigma
        #self.b_mat = reshape(inv((V.transpose())),(1,16))
        #print 'b_mat'
        #print self.b_mat
        return sigma

    def setNPMatrix(self, m_matrix, (h, k, l)):
        """Set fixed NP matrix for a given M matrix at Q point h,k,l."""
        # see below for matrix meanings
        TT = real(self.S*matrix([h, k, l]).transpose())
        cos_theta = TT[0,0]/sqrt(dot(TT.transpose(),TT))
        sin_theta = TT[1,0]/sqrt(dot(TT.transpose(),TT))

        R = [[cos_theta, sin_theta, 0], [-sin_theta, cos_theta, 0], [0, 0, 1]]
        T = matrix(zeros((4, 4)))
        T[3, 3] = 1.
        T[0:3, 0:3] = array(R*self.S).real
        self.NP = T * m_matrix * T.transpose()

    def calcResEllipsoid(self, h, k, l, en):
        # set current Q vector and calculate corresponding resolution matrix
        # (sethklen calls self.calc_popovici() to calculate the matrix...
        if not self.fixed_res:
            self.sethklen(h, k, l, en)
            #   [R0,NP,vi,vf,Error]=feval(method,f,Qmag,p,mon_flag);
            self.R0_corrected = real(self.R0/(sqrt(det(self.NP))/(2*pi)**2))
        # corrected resolution volume as Monte Carlo integral is over a
        # normalised ellipsoid.

        # the resolution matrix calculated by self.calc_popovici is in x, y & z
        # coordinate frame defined by
        #   x || Q
        #   z perp to Q outside scattering plane
        #   y completing r.h.s
        # therefore transformation to h,k,l in (r.l.u.) needs to calculated

        # first calculate angle of Q wrt to V1, V2

        # TT is Q vector in system of V1 and V2 that define scattering plane
        TT = real(self.S*matrix([h, k, l]).transpose())
        # cos(theta) and sin(theta) are the projections of the Q vector onto
        # the directions V1 and V2
        cos_theta = TT[0,0]/sqrt(dot(TT.transpose(),TT))
        sin_theta = TT[1,0]/sqrt(dot(TT.transpose(),TT))

        #----- Rotation matrix from system of resolution matrix
        # to system defined by V1, V2, V3 => V1,V2 define scattering plane,
        # V3 is zone axis, thus perpendicular to the scattering plane
        R = [[cos_theta, sin_theta, 0], [-sin_theta, cos_theta, 0], [0, 0, 1]]
        T = matrix(zeros((4, 4)))
        T[3, 3] = 1.
        T[0:3, 0:3] = array(R*self.S).real
        #self.S performs transformation in h,k,l, en ([Rlu] & [meV])
        self.M = T.transpose()*self.NP*T

    def calcBragg(self):
        """resmat function to calculate the widths (FWHM)
        of a Bragg peak from the resolution matrix M.

        Ported from octave/MATLAB

        Output: bragg, Qx, Qy, Qz, Vanadium and DEE widths
        """
        M = matrix(self.NP)

        bragg = [0, 0, 0, 0, 0]
        bragg[0] = 2.3548/sqrt(M[0,0])
        bragg[1] = 2.3548/sqrt(M[1,1])
        bragg[2] = 2.3548/sqrt(M[2,2])
        _r, bragg[3] = self.calcPhonon(1, matrix([0, 0, 0, 1]))
        bragg[4] = 2.3548/sqrt(M[3,3])
        return abs(bragg)

    def calcPhonon(self,r0,C):
        """resmat  routine to calculate the phonon width of a scan along a
        vector s, and a plane defined by C.X=w. r0 is the resolution constant and
        M is the resolution matrix.

        ported from matlab
        """
        M = matrix(self.NP)
        T = matrix(diag((1, 1, 1, 1)))
        T[3,0:4] = C
        S = inv(T)
        MP = S.transpose()*M*S
        [rp, MP] = rc_int(0, r0, MP)
        [rp, MP] = rc_int(0, rp, MP)
        [rp, MP] = rc_int(0, rp, MP)
        fwhm = 2.35482/sqrt(MP[0,0])
        return [rp, fwhm]

    def showCFG(self):
        """Displays spectrometer configuration."""
        msg =  '\nSpectrometer Configuration:\n'
        msg += '--------------------------\n\n'
        msg += 'NR' + ('Value   :  ').rjust(16) + 'Explanation\n'
        msg += '===============================\n'
        for i in range(len(self.cfg)):
            msg += '%2i' % i + ('%1.3f   :  ' % self.cfg[i]).rjust(16) + \
                '%s\n' % CFGDESC[i]
        print msg

    def resellipse(self):
        """Returns the projections of the resolution ellipse of a triple axis."""
        A = self.NP

        #----- Remove the vertical component from the matrix.
        B = [[A[0,0], A[0,1], A[0,3]],
             [A[1,0], A[1,1], A[1,3]],
             [A[3,0], A[3,1], A[3,3]]]
        B = matrix(B)
        #----- Work out projections for different cuts through the ellipse

        #----- S is the rotation matrix that diagonalises the projected ellipse

        #----- 1. Qx, Qy plane
        # (this is maximal extension of the resolution ellipsoid parallel to x and
        # y, whereas the slice added later is just a cut through the ellipsoid in
        # the xy-plane)
        R0 = self.R0
        _R0P, MP = GaussInt(2, R0, B)
        #print R0P, MP
        hwhm_xp, hwhm_yp, theta = calcEllipseAxis(MP)
        xy_x, xy_y = ellipse_coords(hwhm_xp, hwhm_yp, theta)

        # slice through Qx,Qy plane
        MP = A[0:2,0:2]
        hwhm_xp, hwhm_yp, theta = calcEllipseAxis(MP)
        xys_x, xys_y = ellipse_coords(hwhm_xp, hwhm_yp, theta)

        #----- 2. Qx, W plane
        R0P, MP = GaussInt(1, R0, B)
        hwhm_xp, hwhm_yp, theta = calcEllipseAxis(MP)
        xw_x, xw_y = ellipse_coords(hwhm_xp, hwhm_yp, theta)

        # slice through Qx,W plane
        MP = [[A[0,0], A[0,3]], [A[3,0], A[3,3]]]
        hwhm_xp, hwhm_yp, theta = calcEllipseAxis(MP)
        xws_x, xws_y = ellipse_coords(hwhm_xp, hwhm_yp, theta)

        #----- 3. Qy, W plane
        R0P, MP = GaussInt(0, R0, B)
        hwhm_xp, hwhm_yp, theta = calcEllipseAxis(MP)
        yw_x, yw_y = ellipse_coords(hwhm_xp, hwhm_yp, theta)

        # slice through Qy,W plane
        MP = [[A[1,1], A[1,3]], [A[3,1], A[3,3]]]
        hwhm_xp, hwhm_yp, theta = calcEllipseAxis(MP)
        yws_x, yws_y = ellipse_coords(hwhm_xp, hwhm_yp, theta)

        return xy_x, xy_y, xys_x, xys_y, \
            xw_x, xw_y, xws_x, xws_y, \
            yw_x, yw_y, yws_x, yws_y


def rc_int(index, r0, m):
    """ported from MATLAB

    function that takes a matrix and performs a Gaussian integral
    over the row and column specified by index and returns
    a new matrix. Tested against maple integration.
    """

    r = sqrt(2*pi/m[index,index])*r0

    # remove columns and rows from m
    # that contain the subscript "index".

    mp = m
    b = m[:,index] + m[index,:].transpose()
    c = b[0:len(b)-1]
    c[0:index] = b[0:index]
    c[index:len(b)-1] = b[index+1:len(b)]
    b = c
    T = matrix(zeros((len(mp)-1, len(mp)-1)))
    T[0:index, 0:index] = mp[0:index, 0:index]
    T[index:len(mp)-1, 0:index] = mp[index+1:len(mp), 0:index]
    T[0:index, index:len(mp)-1] = mp[0:index, index+1:len(mp)]
    T[index:(len(mp)-1), index:(len(mp)-1)] = mp[(index+1):len(mp), (index+1):len(mp)]
    #c=[0:len(mp)-1,0:len(mp)-1]
    mp = T
    mp = mp - 1/(4*m[index,index])*b*b.transpose()
    return [r, mp]


# functions for calculating ellipsoid cuts and projections

def calcEllipseAxis(MP):
    const = 1.17741
    MP = matrix(MP)
    theta = 0.5*arctan2(2*MP[0,1], MP[0,0]-MP[1,1])
    S = matrix([[cos(theta), sin(theta)],
                [-sin(theta), cos(theta)]])
    MP = S*MP*S.transpose()
    hwhm_xp = const/sqrt(MP[0,0])
    hwhm_yp = const/sqrt(MP[1,1])
    return hwhm_xp, hwhm_yp, theta

def GaussInt(index, r0, m):
    """Function that takes a matrix and performs a Gaussian integral
    over the row and column specified by index and returns
    a new matrix.

    XXX what is different to rc_int above?
    """
    r = sqrt(2*pi/m[index,index])*r0
    # remove columns and rows from m that contain the subscript "index".
    mp = m
    b = m[:,index]+m[index,:].transpose()
    b = delete(b, index)
    mp = zeros((len(m)-1, len(m)-1))
    mp[0:index, 0:index] = m[0:index, 0:index]
    if index < (len(m)-1):
        mp[index:len(m)-1,0:index] = m[index+1:len(m),0:index]
        mp[0:index,index:len(m)-1] = m[0:index,index+1:len(m)]
        mp[index:len(m)-1,index:len(m)-1] = m[index+1:len(m),index+1:len(m)]
    mp = mp-1./(4*m[index,index])*b.transpose()*b
    return r, mp

def ellipse_coords(a, b, phi):
    """Return coordinates for ellipse with semiaxes a, b and rotated by phi."""
    x0 = 0
    y0 = 0

    th = arange(0, 2*pi+2*pi/100, 2*pi/100)
    x = a*cos(th)
    y = b*sin(th)

    c = cos(phi)
    s = sin(phi)

    th = x*c-y*s+x0
    y = x*s+y*c+y0
    x = th
    return x, y

def single_mc(NMC, sqw, fit_par, QE, b_mat, sigma, R0_corrected):
    xp = zeros((4, NMC))
    xp[0,:] = sigma[0]*randn(NMC)
    xp[1,:] = sigma[1]*randn(NMC)
    xp[2,:] = sigma[2]*randn(NMC)
    xp[3,:] = sigma[3]*randn(NMC)
    XMC = reshape(b_mat[0:16], (4, 4)).transpose() * xp
    XMC = XMC.getA()  # make an array from the matrix

    qh = XMC[0] + QE[0]
    qk = XMC[1] + QE[1]
    ql = XMC[2] + QE[2]
    w  = XMC[3] + QE[3]

    # QE is provided to sqw function in case center of resolution
    # is needed for further calculations
    mc_intens = sqw(qh, qk, ql, w, QE, (b_mat, sigma), *fit_par)
    return R0_corrected * mc_intens.mean()

pool = None

class dummy_result(object):
    def __init__(self, res):
        self.res = res
    def get(self):
        return self.res


def calc_MC_mcstas(x, fit_par, sqw, resmat, NMC):
    """Trial with McStas."""
    results = []
    for QE in x:
        qh, qk, ql, en, weights = resmat.run_mcstas(NMC, QE)
        if qh == []:
            mc_intens = 0
        else:
            mc_intens = (sqw(qh, qk, ql, en, QE, None, *fit_par) * weights).sum()
        results.append(mc_intens)
    return array(results)

def calc_MC(x, fit_par, sqw, resmat, NMC, use_caching=True):
    """Calculates intensity of point in reciprocal space (qh,qk,ql,en) at takes
    into account the spectrometer resolution calculated by resolution class
    resmat (which uses the Popovici algorithm to do so).
    """
    global pool
    if pool is None:
        pool = multiprocessing.Pool(multiprocessing.cpu_count())
    results = []
    for QE in x:
        QE = tuple(QE)
        if QE in resmat._cache and use_caching:
            b_mat, sigma, R0_corrected = resmat._cache[QE]
        else:
            resmat.calcResEllipsoid(*QE)
            if resmat.ERROR:
                print 'Scattering triangle will not close for point: ' \
                    'qh = %1.3f qk = %1.3f ql = %1.3f en = %1.3f' % tuple(QE)
                print 'Attention: Intensity is therefore equal to zero at this point!'
                results.append(dummy_result(0))
                continue
            sigma = resmat.calcSigma()
            b_mat = resmat.b_mat[0:16]
            R0_corrected = resmat.R0_corrected
            resmat._cache[QE] = b_mat, sigma, R0_corrected
        results.append(pool.apply_async(single_mc, (NMC, sqw, fit_par, QE, b_mat,
                                                    sigma, R0_corrected)))
    return array([res.get() for res in results])

single_mc_cluster_code = '''
from numpy import zeros, reshape
from numpy.random import randn
def single_mc(NMC, fit_par, QE, b_mat, sigma, R0_corrected):
    if NMC == 0:
        return 0.

    xp = zeros((4, NMC))
    xp[0,:] = sigma[0]*randn(NMC)
    xp[1,:] = sigma[1]*randn(NMC)
    xp[2,:] = sigma[2]*randn(NMC)
    xp[3,:] = sigma[3]*randn(NMC)
    XMC = reshape(b_mat[0:16], (4, 4)).transpose() * xp
    XMC = XMC.getA()  # make an array from the matrix

    qh = XMC[0] + QE[0]
    qk = XMC[1] + QE[1]
    ql = XMC[2] + QE[2]
    w  = XMC[3] + QE[3]

    # QE is provided to sqw function in case center of resolution
    # is needed for further calculations
    mc_intens = __sqw(qh, qk, ql, w, QE, (b_mat, sigma), *fit_par)
    return R0_corrected * mc_intens.mean()
'''

def calc_MC_cluster(x, fit_par, sqwcode, sqwfunc, resmat, NMC, use_caching=True):
    """Version of calc_MC with clustering support."""
    from ufit import cluster
    args = []
    for QE in x:
        QE = tuple(QE)
        if QE in resmat._cache and use_caching:
            b_mat, sigma, R0_corrected = resmat._cache[QE]
        else:
            resmat.calcResEllipsoid(*QE)
            if resmat.ERROR:
                print 'Scattering triangle will not close for point: ' \
                    'qh = %1.3f qk = %1.3f ql = %1.3f en = %1.3f' % tuple(QE)
                print 'Attention: Intensity is therefore equal to zero at this point!'
                args.append((0, [], None, None, None, None))
                continue
            sigma = resmat.calcSigma()
            b_mat = resmat.b_mat[0:16]
            R0_corrected = resmat.R0_corrected
            resmat._cache[QE] = b_mat, sigma, R0_corrected
        args.append((NMC, fit_par, QE, b_mat, sigma, R0_corrected))
    code = sqwcode + '\n__sqw = %s\n' % sqwfunc + single_mc_cluster_code
    return array(cluster.run_cluster(code, 'single_mc', args))


def load_par(filename):
    pars = {}
    with open(filename) as f:
        for line in f:
            if line.strip():
                parts = line.split()
                name = parts[1].strip('%').lower()
                value = float(parts[0])
                pars[name] = value
    return pars


def load_cfg(filename):
    pars = []
    with open(filename) as f:
        for line in f:
            if line.strip():
                pars.append(float(line.split()[0]))
    return pars


def pylab_key_handler(event):
    import pylab
    if event.key == 'q':
        pylab.close()

def plot_resatpoint(cfg, par, resmat, fignum='Resolution calculation'):
    import pylab

    x, y, xslice, yslice, xxq, yxq, xxqslice, yxqslice, xyq, yyq, xyqslice, yyqslice = \
        resmat.resellipse()

    pylab.figure(fignum, figsize=(8.5, 6), dpi=120, facecolor='1.0')
    pylab.clf()
    pylab.rc('text', usetex=True)
    pylab.rc('text.latex',
             preamble='\\usepackage{amsmath}\\usepackage{helvet}\\usepackage{sfmath}')
    pylab.subplots_adjust(left=0.11, bottom=0.08, right=0.97, top=0.81,
                          wspace=0.25, hspace=0.27)
    # register event handler to pylab
    pylab.connect('key_press_event', pylab_key_handler)

    pylab.subplot(221)
    pylab.xlabel(r'Q$_x$ (\AA$^{-1}$)')
    pylab.ylabel(r'Q$_y$ (\AA$^{-1}$)')
    pylab.plot(x,y)
    pylab.plot(xslice,yslice)

    ax1 = pylab.gca()
    text  = r"""\noindent\underline{Spectrometer Setup:}\newline
\begin{tabular}{ll}
d-spacings: & $d_M=%(dm)1.4f$\,\AA~~~$d_A=%(da)1.4f$\,\AA \\
mosaic:     & $\eta_M=%(etam)3.1f'$~~~$\eta_S=%(etas)3.1f'$~~~$\eta_A=%(etaa)3.1f'$ \\
s-sense:    & $s_M=%(sm)i$~~~$s_S=%(ss)i$~~~$s_A=%(sa)i$ \\
$\alpha_{1\rightarrow4}$: & %(alpha1)i-Mono-%(alpha2)i-Sample-%(alpha3)i-Ana-%(alpha4)i (hor. coll.) \\
$\beta_{1\rightarrow4}$:  & %(beta1)i-Mono-%(beta2)i-Sample-%(beta3)i-Ana-%(beta4)i   (vert. coll.) \\
\end{tabular}
""" % resmat.par
    t1 = pylab.text(-0.25, 1.57, text.replace('\n', ''),
                    horizontalalignment='left', verticalalignment='top',
                    transform=ax1.transAxes)
    t1.set_size(10)

    pylab.subplot(222)
    pylab.xlabel(r'Q$_x$ (\AA$^{-1}$)')
    pylab.ylabel('Energy (meV)')
    pylab.plot(xxq, yxq)
    pylab.plot(xxqslice, yxqslice)

    ax2 = pylab.gca()
    text  = r"""\noindent\underline{Sample Parameters:}\newline
\begin{tabular}{llllll}
$a$ (\AA) & $b$ (\AA) & $c$ (\AA) & $\alpha$ ($^{\circ}$) & $\beta$ ($^{\circ}$) & $\gamma$ ($^{\circ}$) \\
%(as)2.3f & %(bs)2.3f & %(cs)2.3f & %(aa)3.1f & %(bb)3.1f & %(cc)3.1f \\
\end{tabular}\newline """ % resmat.par
    if resmat.par['kfix'] == 1:
        text += r'fixed incident energy $k_i=%2.4f$\,\AA$^{-1}$ ($\equiv %4.2f$\,meV)\newline ' % \
            (resmat.par['k'], resmat.par['k']**2*2.07)
    else:
        text += r'fixed final energy $k_f=%2.4f$\,\AA$^{-1}$ ($\equiv %4.2f$\,meV)\newline ' % \
            (resmat.par['k'], resmat.par['k']**2*2.07)
    text += r'position: qh = %1.3f qk = %1.3f ql = %1.3f (r.l.u.) en = %2.3f (meV)\newline ' % \
        (resmat.par['qx'], resmat.par['qy'], resmat.par['qz'], resmat.par['en'])
    text += r'modulus of scattering vector $Q = %2.5f$\,\AA$^{-1}$' % resmat.q0
    t2 = pylab.text(-0.25, 1.57, text.replace('\n', ' '),
                    horizontalalignment='left', verticalalignment='top',
                    transform=ax2.transAxes)
    t2.set_size(10)

    pylab.subplot(223)
    pylab.xlabel(r'Q$_y$ (\AA$^{-1}$)')
    pylab.ylabel(r'Energy (meV)')
    pylab.plot(xyq,yyq)
    pylab.plot(xyqslice,yyqslice)

    pylab.subplot(224)
    ax3 = pylab.gca()
    pylab.axis('off')
    pylab.rc('text', usetex=True)
    text  = r'\noindent\underline{\textbf{Resolution Info:}}\newline ' + \
        r'Resolution Volume: $R_0 = %7.5g$ (\AA$^{-3}$\,meV)\newline\newline ' % resmat.R0
    mat = resmat.NP.tolist()
    text += r'Resolution Matrix (in frame $Q_x$, $Q_y$, $Q_z$, $E$):\newline '
    text += (r'$M = \left(\begin{array}{rrrr} %5.2f & %5.2f & %5.2f & %5.2f\\ ' \
             r'%5.2f & %5.2f & %5.2f & %5.2f\\ %5.2f & %5.2f & %5.2f & %5.2f\\ ' \
             r'%5.2f & %5.2f & %5.2f & %5.2f \end{array}\right)$\newline\newline\newline ') \
             % (mat[0][0], mat[0][1], mat[0][2], mat[0][3],
                mat[1][0], mat[1][1], mat[1][2], mat[1][3],
                mat[2][0], mat[2][1], mat[2][2], mat[2][3],
                mat[3][0], mat[3][1], mat[3][2], mat[3][3])
    text += r'Bragg width:\newline '
    text += r'\begin{tabular}{lllll} '
    text += r'$Q_x$ (\AA$^{-1}$) & $Q_y$ (\AA$^{-1}$) & $Q_z$ (\AA$^{-1}$) & Vanadium & dE (meV) \\ '
    bragw = tuple(resmat.calcBragg())
    text += r'%1.5f & %1.5f & %1.5f & %1.5f & %1.5f \\ ' % bragw[:5]
    text += r'\end{tabular}'
    t3 = pylab.text(-0.13, 1.0, text,
                    horizontalalignment='left', verticalalignment='top',
                    transform=ax3.transAxes)
    t3.set_size(10)
    pylab.show()



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Session and session item abstraction."""

from os import path
import cPickle as pickle

from PyQt4.QtCore import QObject, SIGNAL

from ufit import UFitError
from ufit.utils import attrdict


# current save file version
SAVE_VERSION = 3


class SessionItem(QObject):
    """Represents an "item" in the item list of a session."""

    def __init__(self):
        QObject.__init__(self)
        self.group = None
        self.index = -1
        self.title = ''
        self.htmldesc = ''

    def set_group(self, group, index):
        self.group = group
        self.index = index
        self.update_htmldesc()

    def after_load(self):
        """Do whatever is necessary to upgrade the data format."""

    def update_htmldesc(self):
        """Update self.htmldesc and self.title."""
        self.htmldesc = self.title = '(item)'

    def create_panel(self, mainwindow, canvas):
        """Create a GUI panel that represents this item."""
        raise NotImplementedError('implement create_panel')

    def __reduce__(self):
        """Need a special __reduce__ since we don't want to pickle the panel."""
        raise NotImplementedError('implement __reduce__')

    def export_python(self, filename):
        """Export item as Python code."""

    def export_ascii(self, filename):
        """Export item data as ASCII data."""

    def export_fits(self, filename):
        """Export fits as ASCII data."""


class ItemGroup(object):
    expanded = False

    def __init__(self, name):
        self.name = name
        self.items = []
        self.update_htmldesc()

    def update_htmldesc(self):
        nitems = len(self.items)
        self.htmldesc = '<img src=":/drawer.png">&nbsp;&nbsp;<b>%s</b>' \
                        ' &ndash; %d items' % (self.name, nitems)


class _Session(QObject):

    def __init__(self):
        QObject.__init__(self)
        self.filename = None
        self.groups = []
        self.all_items = set()
        self.clear()

    @property
    def dirname(self):
        if self.filename:
            return path.dirname(self.filename)
        return ''

    def clear(self):
        self.groups[:] = [ItemGroup('Default')]
        self.all_items.clear()
        self.props = attrdict()
        self.emit(SIGNAL('itemsUpdated'))
        self.emit(SIGNAL('propsUpdated'))
        self.emit(SIGNAL('dirtyChanged'), False)

    def new(self):
        self.filename = None
        self.emit(SIGNAL('filenameChanged'))
        self.clear()

    def set_filename(self, filename):
        self.filename = filename
        self.emit(SIGNAL('filenameChanged'))

    def _load_v0(self, info):
        info['version'] = 1
        info['datasets'] = info.pop('panels')
        info['template'] = ''
        self._load_v1(info)

    def _load_v1(self, info):
        info['version'] = 2
        datasets = info.pop('datasets')
        info['panels'] = [('dataset', d[0], d[1]) for d in datasets]
        self._load_v2(info)

    def _load_v2(self, info):
        info['version'] = 3
        group = ItemGroup('Default')
        info['groups'] = [group]
        panels = info.pop('panels')
        from ufit.gui.mappingitem import MappingItem
        from ufit.gui.scanitem import ScanDataItem
        for panel in panels:
            if panel[0] == 'dataset':
                group.items.append(ScanDataItem(panel[1], panel[2]))
            elif panel[0] == 'mapping':
                group.items.append(MappingItem(panel[1], panel[2]))
        info['props'] = attrdict()
        info['props'].template = info.pop('template')
        self._load_v3(info)

    def _load_v3(self, info):
        self.props = info['props']
        self.groups[:] = info['groups']

    def load(self, filename):
        self.clear()
        # unpickle everything
        with open(filename, 'rb') as fp:
            info = pickle.load(fp)
        # load with the respective method
        savever = info.get('version', 0)
        try:
            getattr(self, '_load_v%d' % savever)(info)
        except AttributeError:
            raise UFitError('save version %d not supported' % savever)
        self.filename = filename
        # reassign indices (also to regenerate descriptions)
        for group in self.groups:
            for i, item in enumerate(group.items):
                item.set_group(group, i + 1)
                item.after_load()
                self.all_items.add(item)
            group.update_htmldesc()
        # let GUI elements update from propsdata
        self.emit(SIGNAL('itemsUpdated'))
        self.emit(SIGNAL('propsUpdated'))
        self.emit(SIGNAL('filenameChanged'))

    def save(self):
        # let GUI elements update the stored propsdata
        self.emit(SIGNAL('propsRequested'))
        if self.filename is None:
            raise UFitError('session has no filename yet')
        info = {
            'version':  SAVE_VERSION,
            'groups':   self.groups,
            'props':    self.props,
        }
        with open(self.filename, 'wb') as fp:
            pickle.dump(info, fp, protocol=pickle.HIGHEST_PROTOCOL)
        self.emit(SIGNAL('dirtyChanged'), False)

    def add_group(self, name):
        group = ItemGroup(name)
        self.groups.append(group)
        self.set_dirty()
        self.emit(SIGNAL('groupAdded'), group)
        return group

    def remove_group(self, group):
        self.groups.remove(group)
        for item in group.items:
            self.all_items.discard(item)
        self.set_dirty()
        self.emit(SIGNAL('itemsUpdated'))

    def rename_group(self, group, name):
        group.name = name
        group.update_htmldesc()
        self.set_dirty()
        self.emit(SIGNAL('groupUpdated'), group)

    def add_item(self, item, group=None):
        if group is None:
            group = self.groups[-1]
        self.all_items.add(item)
        group.items.append(item)
        group.update_htmldesc()
        item.set_group(group, len(group.items))
        self.set_dirty()
        self.emit(SIGNAL('itemAdded'), item)

    def add_items(self, items, group=None):
        if not items:
            return
        if group is None:
            group = self.groups[-1]
        self.all_items.update(items)
        for item in items:
            group.items.append(item)
            item.set_group(group, len(group.items))
        group.update_htmldesc()
        self.set_dirty()
        self.emit(SIGNAL('itemsUpdated'))
        self.emit(SIGNAL('itemAdded'), items[-1])

    def remove_items(self, items):
        renumber_groups = set()
        for item in items:
            renumber_groups.add(item.group)
            item.group.items.remove(item)
            self.all_items.discard(item)
        for group in renumber_groups:
            if not group.items:
                self.groups.remove(group)
            for i, item in enumerate(group.items):
                item.set_group(group, i + 1)
            group.update_htmldesc()
        self.set_dirty()
        self.emit(SIGNAL('itemsUpdated'))

    def move_items(self, items, newgroup):
        renumber_groups = set([newgroup])
        for item in items:
            renumber_groups.add(item.group)
            item.group.items.remove(item)
            newgroup.items.append(item)
        for group in renumber_groups:
            for i, item in enumerate(group.items):
                item.set_group(group, i + 1)
            group.update_htmldesc()
        self.set_dirty()
        self.emit(SIGNAL('itemsUpdated'))

    def reorder_groups(self, new_structure):
        del self.groups[:]
        for group, items in new_structure:
            self.groups.append(group)
            group.items[:] = items
            for i, item in enumerate(items):
                item.set_group(group, i + 1)
            group.update_htmldesc()
        self.set_dirty()
        self.emit(SIGNAL('itemsUpdated'))

    def set_dirty(self):
        self.emit(SIGNAL('dirtyChanged'), True)

# one singleton instance
session = _Session()

# one temporary instance
temp_session = _Session()



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Various data display widgets for the GUI."""

from PyQt4.QtCore import Qt
from PyQt4.QtGui import QListWidget, QListWidgetItem


class DataValueListWidget(QListWidget):
    """View of data values from some datasets."""

    def populate(self, items, intersect=True):
        allparams = set()
        alldatav = set()
        for item in items:
            if not item.model or not item.data:
                return
            pgen = (p.name for p in item.model.params)
            if not allparams:
                allparams = set(pgen)
            elif intersect:
                allparams.intersection_update(pgen)
            else:
                allparams.update(pgen)
            vgen = (v for v in item.data.meta if not v.startswith('col_'))
            if not alldatav:
                alldatav = set(vgen)
            elif intersect:
                alldatav.intersection_update(vgen)
            else:
                alldatav.update(vgen)

        wi = QListWidgetItem('Parameters', self)
        wi.setFlags(Qt.NoItemFlags)  # not selectable
        for param in sorted(allparams):
            QListWidgetItem('   ' + param, self, 1)
        wi = QListWidgetItem('Data values', self)
        wi.setFlags(Qt.NoItemFlags)
        for datav in sorted(alldatav):
            QListWidgetItem('   ' + datav, self, 2)



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Diverse other models."""

from numpy import cos, sin, exp, pi, log, piecewise, sign, tanh

from ufit.param import Param
from ufit.models.base import Model

__all__ = ['Const', 'StraightLine', 'Parabola',
           'Cosine', 'Sinc', 'ExpDecay', 'PowerLaw', 'BrillouinT', 'BrillouinB']


class Const(Model):
    """A constant, to be used for modifying other models (e.g. exponentiation)

    For example: ``Sinc() ** Const('eta')``

    Parameters:

    * the constant  (named after the model's name)
    """

    def __init__(self, name='', c=None):
        # this initialization is a bit different from what _init_params does,
        # so we don't use it here: the parameter's name will be whatever the
        # model itself is called
        self.name = name
        pname = name or 'c'
        self.params = [Param.from_init(pname, c or 0)]
        self.fcn = lambda p, x: p[pname] + 0*x

    def is_modifier(self):
        return True


class StraightLine(Model):
    """Straight line

    y = slope * x + y0

    Parameters:

    * `slope`
    * `y0`  - intercept
    """
    param_names = ['slope', 'y0']

    def __init__(self, name='', slope=1, y0=0):
        psl, py0 = self._init_params(name, self.param_names, locals())
        self.fcn = lambda p, x: p[psl]*x + p[py0]

    pick_points = ['one point on curve', 'another point on curve']

    def convert_pick(self, b1, b2):
        slope = (b2[1] - b1[1]) / (b2[0] - b1[0])
        return {
            self.params[0].name: slope,
            self.params[1].name: b1[1] - slope*b1[0],
        }


class Parabola(Model):
    """Parabola

    y = stretch * (x - x0)^2 + y0

    Parameters:

    * `x0`  - x coordinate of vertex
    * `y0`  - y coordinate of vertex
    * `stretch` - stretch factor
    """
    param_names = ['x0', 'y0', 'stretch']

    def __init__(self, name='', x0=0, y0=0, stretch=1):
        px0, py0, ps = self._init_params(name, self.param_names, locals())
        self.fcn = lambda p, x: p[ps] * (x - p[px0])**2 + p[py0]

    pick_points = ['vertex', 'another point on curve']

    def convert_pick(self, vx, p2):
        return {
            self.params[0].name: vx[0],
            self.params[1].name: vx[1],
            self.params[2].name: (p2[1] - vx[1]) / (p2[0] - vx[0])**2,
        }



class Cosine(Model):
    """Cosine

    y = ampl * cos(freq * x + phase)

    Parameters:

    * `ampl` - amplitude
    * `freq` - frequency (omega or k)
    * `phase` - phase in radians
    """
    param_names = ['ampl', 'freq', 'phase']

    def __init__(self, name='', ampl=None, freq=None, phase=0):
        pa, pf, pp = self._init_params(name, self.param_names, locals())
        self.fcn = lambda p, x: p[pa] * cos(p[pf]*x + p[pp])

    pick_points = ['a maximum', 'next minimum']

    def convert_pick(self, pmax, pmin):
        freq = pi/abs(pmin[0] - pmax[0])
        return {
            self.params[0].name: pmax[1] - pmin[1],          # amplitude
            self.params[1].name: freq,                       # frequency
            self.params[2].name: (- freq*pmax[0]) % (2*pi),  # phase
        }


class Sinc(Model):
    """Sinc function

    y = ampl * sin(freq*(x - center)) / (freq*(x - center))

    Parameters:

    * `ampl` - amplitude at x = center
    * `freq` - frequency of the sine
    * `center` - point of maximum amplitude
    """
    param_names = ['ampl', 'freq', 'center']

    def __init__(self, name='', ampl=None, freq=None, center=0):
        pa, pf, pc = self._init_params(name, self.param_names, locals())
        self.fcn = lambda p, x: piecewise(
            x - p[pc], [x == p[pc]],
            [p[pa], lambda v: p[pa] * sin(p[pf]*v) / (p[pf]*v)])


class ExpDecay(Model):
    """Exponential decay

    y = y1 + (y0 - y1) * exp(-x/tau)

    Parameters:

    * `y0`  - value at x = 0
    * `tau` - decay constant: exp(-x/tau)
    * `y1`  - value at x -> infinity
    """
    param_names = ['y0', 'tau', 'y1']

    def __init__(self, name='', y0=1, tau=None, y1=0):
        p0, pt, p1 = self._init_params(name, self.param_names, locals())
        self.fcn = lambda p, x: p[p1] + (p[p0]-p[p1])*exp(-x/p[pt])

    pick_points = ['maximum value', 'minimum value', 'half maximum']

    def convert_pick(self, pmax, pmin, phalf):
        return {
            self.params[0].name: pmax[1],
            self.params[1].name: phalf[0]/log(2),
            self.params[2].name: pmin[1],
        }


class PowerLaw(Model):
    """Power law

    Parameters:

    * `start` - starting point
    * `scale` - x value scaling (positive => right side of starting point)
    * `beta`  - exponent
    """
    param_names = ['start', 'scale', 'beta']

    def __init__(self, name='', scale=1, start=0, ampl=None, beta=None):
        ps, psc, pb = self._init_params(name, self.param_names, locals())
        self.fcn = lambda p, x: piecewise(
            p[psc]*(x-p[ps]), [p[psc]*(x-p[ps]) < 0],
            [0, lambda v: pow(v, p[pb])])

    pick_points = ['starting point', 'one point on curve',
                   'another point on curve']

    def convert_pick(self, pstart, p1, p2):
        beta = log(p1[1]/p2[1])/log((p1[0]-pstart[0])/(p2[0]-pstart[0]))
        scale = sign(p1[0] - pstart[0])*(p1[1] / abs(p1[0] - pstart[0])**beta)
        return {
            self.params[0].name: pstart[0],
            self.params[1].name: scale,
            self.params[2].name: beta,
        }


class BrillouinT(Model):
    """Brillouin function versus temperature (in K)

    Parameters:

    * `J` - spin
    * `B` - applied field in T
    * `g` - g-factor
    * `scale` - scale of the Y values
    """
    param_names = ['J', 'B', 'g', 'scale']

    def __init__(self, name='', J=1, B=0, g=1, scale=1):
        pj, pb, pg, ps = self._init_params(name, self.param_names, locals())
        def fcn(p, x):
            J = p[pj]
            arg = 0.67171388 * p[pg] * p[pb] / x
            return p[ps] * ((2*J+1)/(2*J)/tanh((2*J+1)/(2*J)*arg) -
                            1/(2*J)/tanh(arg/(2*J)))
        self.fcn = fcn


class BrillouinB(Model):
    """Brillouin function versus field (in T)

    Parameters:

    * `J` - spin
    * `T` - temperature in K
    * `g` - g-factor
    * `scale` - scale of the Y values
    """
    param_names = ['J', 'T', 'g', 'scale']

    def __init__(self, name='', J=1, T=1, g=1, scale=1):
        pj, pt, pg, ps = self._init_params(name, self.param_names, locals())
        def fcn(p, x):
            J = p[pj]
            arg = 0.67171388 * p[pg] * x / p[pt]
            return p[ps] * ((2*J+1)/(2*J)/tanh((2*J+1)/(2*J)*arg) -
                            1/(2*J)/tanh(arg/(2*J)))
        self.fcn = fcn



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Session annotation window."""

from PyQt4.QtCore import SIGNAL, QByteArray
from PyQt4.QtGui import QMainWindow, QTextEdit

from ufit.gui.common import SettingGroup, loadUi
from ufit.gui.session import session


class AnnotationWindow(QMainWindow):
    def __init__(self, parent):
        self._updating = False
        self._data = None
        QMainWindow.__init__(self, parent)
        loadUi(self, 'annotations.ui')
        self.sgroup = SettingGroup('annotations')

        self.connect(session, SIGNAL('propsRequested'),
                     self.on_session_propsRequested)
        self.connect(session, SIGNAL('propsUpdated'),
                     self.on_session_propsUpdated)

        with self.sgroup as settings:
            geometry = settings.value('geometry', QByteArray())
            self.restoreGeometry(geometry)
            windowstate = settings.value('windowstate', QByteArray())
            self.restoreState(windowstate)

        self._editing = True
        self.on_session_propsUpdated()

    def on_textBox_textChanged(self):
        if self._editing:
            session.set_dirty()

    def on_session_propsRequested(self):
        session.props.annotations = self.textBox.toPlainText()

    def on_session_propsUpdated(self):
        if 'annotations' in session.props:
            self._editing = False
            self.textBox.setText(session.props.annotations)
            self._editing = True

    def closeEvent(self, event):
        self.on_session_propsRequested()
        with self.sgroup as settings:
            settings.setValue('geometry', self.saveGeometry())
            settings.setValue('windowstate', self.saveState())
        self.emit(SIGNAL('closed'))
        return QMainWindow.closeEvent(self, event)



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Base dataset class."""

import copy
import operator

from numpy import array, concatenate, ones, broadcast_arrays, savetxt, sqrt

from ufit.utils import attrdict
from ufit.data.merge import rebin, floatmerge
from ufit.plotting import DataPlotter


def sanitize_meta(meta, name):
    """
    Sanitize metadata so that keys that must always be there, are there.

    These special metadata keys, to be set by the loaders if possible, are:

    * instrument: the name of the instrument used
    * experiment: the number or short name of the experiment/proposal
    * filenumber: the file number of the data file
    * filedesc: a combination of instrument, experiment and filenumber
    * title: "title" of the data file, i.e. experiment name or sample name
    * subtitle: "scan info", normally the scan command used
    * environment: a list of "sample environment" type strings
    * datafilename: the file name when the data was loaded
    """
    if 'instrument' not in meta:
        meta.instrument = ''
    if 'experiment' not in meta:
        meta.experiment = ''
    if 'filenumber' not in meta:
        meta.filenumber = 0
    if 'title' not in meta:
        meta.title = name or str(meta.filenumber or '---')
    if 'subtitle' not in meta:
        # 'info' is the old name of 'subtitle'
        meta.subtitle = meta.get('info', '')
    if 'environment' not in meta:
        meta.environment = []
    if 'filedesc' not in meta:
        meta.filedesc = '%s:%s:%s' % (meta.instrument, meta.experiment,
                                      meta.filenumber)
    if 'datafilename' not in meta:
        meta.datafilename = ''


class DataBase(object):

    def __init__(self, meta, name='', sources=None):
        self.meta = attrdict(meta)
        sanitize_meta(self.meta, name)
        self.name = name or str(self.meta.filenumber or '---')
        self.sources = sources or [self.meta.filedesc]

    def after_load(self):
        pass

    def __getattr__(self, key):
        if key == '__setstate__':
            # pickling support
            raise AttributeError(key)
        elif key in self.meta:
            return self.meta[key]
        elif key == 'x_plot':
            # backwards compatibility
            self.x_plot = self.x
            return self.x
        raise AttributeError('no such data column or metadata: %s' % key)

    def copy(self):
        return copy.deepcopy(self)


class ScanData(DataBase):
    def __init__(self, meta, data, xcol, ycol, ncol=None, nscale=1,
                 name='', sources=None):
        DataBase.__init__(self, meta, name, sources)

        self._data = data

        self.xcol = self.xaxis = xcol
        self.x = self.x_raw = data[:,0]
        self.x_plot = self.x

        self.ycol = self.yaxis = ycol
        self.y_raw = data[:,1]
        self.dy_raw = data[:,2]

        self.ncol = ncol
        self.nscale = nscale
        if ncol is not None and data.shape[1] > 3:
            self.norm_raw = data[:,3]
            self.norm = self.norm_raw / nscale
            if nscale != 1:
                self.yaxis += ' per %s %s' % (nscale, ncol)
            else:
                self.yaxis += ' / %s' % ncol
        else:
            self.norm_raw = self.norm = ones(len(self.y_raw))

        self.y = self.y_raw / self.norm
        self.dy = self.dy_raw / self.norm

        self.reset_mask()
        self.fitmin = None
        self.fitmax = None

    def after_load(self):
        """Update internal data structures after unpickling."""
        if 'fitmin' not in self.__dict__:
            self.fitmin = self.fitmax = None
        if 'mask' not in self.__dict__:
            self.reset_mask()
        sanitize_meta(self.meta, self.name)

    @property
    def fit_columns(self):
        mask = self.mask.copy()
        if self.fitmin is not None:
            mask &= self.x >= self.fitmin
        if self.fitmax is not None:
            mask &= self.x <= self.fitmax
        return self.x[mask], self.y[mask], self.dy[mask]

    @classmethod
    def from_arrays(cls, name, x, y, dy, meta=None, xcol='x', ycol='y'):
        multix = False
        if isinstance(x, list):
            x = array(x)
        if len(x.shape) > 1:
            multix = True
            xorig = x
            x = x[:,0]
        data = array(broadcast_arrays(x, y, dy)).T
        ret = cls(meta or {}, data, xcol, ycol, name=name)
        if multix:
            ret.x = xorig
        return ret

    def reset_mask(self):
        # points with mask = False are masked out
        self.mask = ones(len(self.x), bool)
        self.mask[self.dy==0] = False

    def __repr__(self):
        return '<%s (%d points)>' % (self.name, len(self.x))

    def __getitem__(self, key):
        if isinstance(key, slice):
            return self.__class__(self.meta, self._data[key],
                                  self.xcol, self.ycol, self.ncol, self.nscale,
                                  name=self.name, sources=self.sources)
        return getattr(self, key)

    def __or__(self, other):
        return self.__class__(self.meta, concatenate((self._data, other._data)),
                              self.xcol, self.ycol, self.ncol, self.nscale,
                              name=self.name + '|' + other.name,
                              sources=self.sources + other.sources)

    def merge(self, binsize, *others, **settings):
        """Merge this dataset with others.

        The X values are redistributed into bins according to the given bin
        size.
        """
        if not others and binsize == 0:
            return self
        allsets = (self,) + others
        alldata = concatenate([dset._data for dset in allsets])
        if settings.get('floatmerge'):
            new_array = floatmerge(alldata, binsize)
        else:
            new_array = rebin(alldata, binsize)
        sources = sum((dset.sources for dset in allsets), [])
        new_meta = self.meta.copy()
        for col in self.meta:
            if not col.startswith('col_'):
                continue
            new_meta[col] = concatenate([dset.meta.get(col, []) for dset in allsets])
        # XXX hkl data is a mess
        if 'is_hkldata' in self.meta:
            concat = concatenate([dset.meta['hkle'] for dset in allsets])
            if binsize != 0:
                if len(set(dset.meta['hkle_vary'] for dset in allsets)) != 1:
                    raise Exception('datasets have differing varying dimension')
                #print concat
                concat = array([concat[0]]*len(new_array))
                concat[:,['h', 'k', 'l', 'E'].index(self.meta['hkle_vary'])] = new_array[:,0]
            #print concat
            new_meta['hkle'] = concat
        # XXX should we merge other meta's?
        ret = self.__class__(new_meta, new_array,
                             self.xcol, self.ycol, self.ncol, self.nscale,
                             name='_'.join(d.name for d in allsets),
                             sources=sources)
        if 'is_hkldata' in self.meta:
            ret.x = ret.meta['hkle']
        return ret

    def plot(self, axes=None, symbols=True, lines=False, **kw):
        """Plot the dataset using matplotlib.

        *axes* is a matplotlib Axes object, as returned by :func:`gca()`.  If
        no axes are given, the current figure is used.

        *symbols* and *lines* control whether the data is plotted using symbols,
        lines or both.
        """
        dp = DataPlotter(axes=axes)
        dp.symbols = symbols
        dp.lines = lines
        dp.plot_data(self, **kw)

    def export_ascii(self, fp):
        savetxt(fp, array([self.x, self.y, self.dy]).T)

    def export_python(self, fp, objname='data'):
        fp.write('%s = as_data(%r, %r, %r, %r)\n' %
                 (objname, self.x, self.y, self.dy, self.name))

# compatibility name
Dataset = ScanData


class ImageData(DataBase):
    def __init__(self, meta, arr, darr, norm=None, nscale=1,
                 name='', sources=None):
        DataBase.__init__(self, meta, name, sources)

        self.arr_raw = arr
        self.darr_raw = darr
        self.nscale = nscale

        if norm:
            self.norm_raw = norm
            self.norm = norm / nscale
        else:
            self.norm_raw = self.norm = 1

        self.arr = arr / self.norm
        self.darr = darr / self.norm

        # XXX implement scaling?
        self.xaxis = 'pixels X'
        self.yaxis = 'pixels Y'

    def __reduce__(self):
        # avoid storing both arr and arr_raw in the pickle files
        return (self.__class__, (self.meta, self.arr, self.darr, self.norm,
                                 self.nscale, self.name, self.sources))

    def __add__(self, other):
        if not isinstance(other, ImageData):
            raise TypeError
        return self.__class__(self.meta, self.arr_raw + other.arr_raw,
                              sqrt(self.darr_raw**2 + other.darr_raw**2),
                              self.norm_raw + other.norm_raw,
                              self.nscale, name=self.name + '+' + other.name,
                              sources=self.sources + other.sources)

    def __subtract__(self, other):
        if not isinstance(other, ImageData):
            raise TypeError
        scaled_arr = other.arr / other.nscale * self.nscale
        scaled_darr = other.darr / other.nscale * self.nscale
        return self.__class__(self.meta, self.arr_raw - scaled_arr,
                              sqrt(self.darr_raw**2 + scaled_darr**2),
                              self.norm_raw, self.nscale,
                              name=self.name + '-' + other.name,
                              sources=self.sources + other.sources)

    def plot(self, axes=None, **kw):
        """Plot the image dataset using matplotlib.

        *axes* is a matplotlib Axes object, as returned by :func:`gca()`.  If
        no axes are given, the current figure is used.
        """
        dp = DataPlotter(axes=axes)
        dp.plot_image(self, **kw)

    def merge(self, binsize, *others, **kwds):
        return reduce(operator.add, others, self)

    def __repr__(self):
        return '<%s (%dx%d pixels)>' % (self.name, self.arr.shape[0],
                                        self.arr.shape[1])


class DataList(dict):

    def __getitem__(self, obj):
        if isinstance(obj, slice):
            return [self[i] for i in range(*obj.indices(10000))]
        return dict.__getitem__(self, obj)

    def c(self, r1, r2):
        return reduce(lambda a, b: a|b, self[r1:r2])


class DatasetList(list):

    def __getattr__(self, key):
        return array([getattr(d, key) for d in self])



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Plotting routines for ufit using matplotlib."""

from itertools import cycle

from numpy import array, isscalar, linspace

import matplotlib
matplotlib.rc('font', family='Helvetica')
matplotlib.rc('savefig', format='pdf')

from matplotlib import pyplot as pl
from matplotlib.colors import LogNorm

from ufit.param import prepare_params


def multi_linspace(start, stop, steps):
    """linspace variant that handles arrays:

    multi_linspace([0,5], [1,10], 3) gives [[0,5],[0.5,7.5],[1,10]].
    """
    if isscalar(start):
        return linspace(start, stop, steps)
    return array([linspace(start[i], stop[i], steps)
                  for i in range(len(start))]).T


class DataPlotter(object):

    markers = ['o', 's', 'v', '*', 'd', '^', '>', 'p']

    def __init__(self, canvas=None, axes=None, toolbar=None):
        if axes is None:
            axes = pl.gca()
        self.axes = axes
        self.orig_spspec = axes.get_subplotspec()
        self.canvas = canvas
        self.image = None
        self.save_layout()
        self.marker_cycle = cycle(self.markers)
        self.toolbar = toolbar
        self._limits = None
        # display options
        self.no_fits = False
        self.symbols = True
        self.lines = False
        self.grid = True
        self.legend = True
        self.imgsmoothing = True

    def draw(self):
        self.canvas.draw()

    def save_layout(self):
        self.orig_axes_position = self.axes.get_position()

    def reset(self, limits=None):
        if limits is True:
            self._limits = self.axes.get_xlim(), self.axes.get_ylim()
        else:
            self._limits = limits
        xscale, yscale = self.axes.get_xscale(), self.axes.get_yscale()
        if self.image is not None:
            self.canvas.figure.delaxes(self.image.colorbar.ax)
            self.image = None
            self.axes.set_subplotspec(self.orig_spspec)
            self.axes.set_position(self.orig_axes_position)
        self.axes.clear()
        self.axes.set_aspect('auto')
        self.axes.set_xscale(xscale)
        self.axes.set_yscale(yscale)
        self.marker_cycle = cycle(self.markers)

    def plot_data(self, data, multi=False, ms=8, offset=0, **kw):
        """Plot dataset."""
        axes = self.axes
        marker = self.marker_cycle.next() if self.symbols else ''
        ls = '-' if self.lines else ''
        if 'label' not in kw:
            kw['label'] = data.name
        if data.mask.all():
            eb = axes.errorbar(data.x_plot, data.y + offset, data.dy, ls=ls,
                               marker=marker, ms=ms, picker=5, **kw)
            color = eb[0].get_color()
        else:
            mask = data.mask
            eb = axes.errorbar(data.x_plot[mask], data.y[mask] + offset,
                               data.dy[mask], ls=ls, marker=marker, ms=ms,
                               picker=5, **kw)
            color = eb[0].get_color()
            kw['label'] = ''
            axes.errorbar(data.x_plot[~mask], data.y[~mask] + offset,
                          data.dy[~mask], ls='', marker=marker, ms=ms,
                          picker=5, mfc='white', mec=color, **kw)
        if not multi:
            if data.fitmin is not None:
                axes.axvline(data.fitmin, ls='-', color='gray')
            if data.fitmax is not None:
                axes.axvline(data.fitmax, ls='-', color='grey')
            axes.set_title('%s\n%s' % (data.title, data.subtitle),
                           size='medium')
            self.plot_finish(data.xaxis, data.yaxis)
        return color

    def plot_finish(self, xlabel=None, ylabel=None, title=None):
        axes = self.axes
        if xlabel is not None:
            axes.set_xlabel(xlabel)
        if ylabel is not None:
            axes.set_ylabel(ylabel)
        if title is not None:
            axes.set_title(title, size='medium')
        if self.legend:
            axes.legend(prop={'size': 'small'})
        if self.grid:
            axes.grid(True)
        if self._limits:
            axes.set_xlim(*self._limits[0])
            axes.set_ylim(*self._limits[1])
        if self.toolbar:
            self.toolbar.update()

    def _get_samples(self, model, data):
        if model.nsamples < 0:
            return len(data.x) * (-model.nsamples)
        return model.nsamples

    def plot_model_full(self, model, data, labels=True, paramvalues=None,
                        offset=0, **kw):
        if self.no_fits:
            return
        if paramvalues is None:
            paramvalues = prepare_params(model.params, data.meta)[3]
        nsamples = self._get_samples(model, data)
        imin, imax = data.x_plot.argmin(), data.x_plot.argmax()
        xx = multi_linspace(data.x[imin], data.x[imax], nsamples)
        xxp = linspace(data.x_plot[imin], data.x_plot[imax], nsamples)
        yy = model.fcn(paramvalues, xx)
        if 'label' not in kw:
            kw['label'] = labels and 'fit' or ''
        self.axes.plot(xxp, yy + offset, 'g', lw=kw.pop('kw', 2), **kw)
        for comp in model.get_components():
            if comp is model:
                continue
            yy = comp.fcn(paramvalues, xx)
            kw['label'] = labels and comp.name or ''
            self.axes.plot(xxp, yy + offset, '-.', **kw)

    def plot_model(self, model, data, labels=True, paramvalues=None,
                   offset=0, **kw):
        if self.no_fits:
            return
        if paramvalues is None:
            paramvalues = prepare_params(model.params, data.meta)[3]
        nsamples = self._get_samples(model, data)
        imin, imax = data.x_plot.argmin(), data.x_plot.argmax()
        xx = multi_linspace(data.x[imin], data.x[imax], nsamples)
        xxp = linspace(data.x_plot[imin], data.x_plot[imax], nsamples)
        yy = model.fcn(paramvalues, xx)
        if 'label' not in kw:
            kw['label'] = labels and 'fit' or ''
        self.axes.plot(xxp, yy + offset, kw.pop('fmt', 'g'),
                       lw=kw.pop('lw', 2), **kw)

    def plot_model_components(self, model, data, labels=True, paramvalues=None,
                              offset=0, **kw):
        if self.no_fits:
            return
        if paramvalues is None:
            paramvalues = prepare_params(model.params, data.meta)[3]
        nsamples = self._get_samples(model, data)
        imin, imax = data.x_plot.argmin(), data.x_plot.argmax()
        xx = multi_linspace(data.x[imin], data.x[imax], nsamples)
        xxp = linspace(data.x_plot[imin], data.x_plot[imax], nsamples)
        for comp in model.get_components():
            yy = comp.fcn(paramvalues, xx)
            kw['label'] = labels and comp.name or ''
            self.axes.plot(xxp, yy + offset, kw.pop('fmt', '-.'), **kw)

    def plot_params(self, params, chisqr):
        s = []
        for p in params:
            if p.expr:
                continue
            s.append(u'%-12s = %9.4g ± %9.4g' % (p.name, p.value, p.error))
            if p.expr:
                s[-1] += ' (fixed)'
            if p.overall:
                s[-1] += ' (global)'
        s.append(u'%-12s = %9.3f' % (u'chi²/NDF', chisqr))
        s = '\n'.join(s)
        self.axes.text(0.02, 0.98, s, horizontalalignment='left',
                       verticalalignment='top', size='x-small',
                       transform=self.axes.transAxes, family='Monospace')

    def plot_mapping(self, *args, **kwds):
        kwds['axes'] = self.axes
        kwds['figure'] = self.canvas.figure
        kwds['clear'] = False
        self.image = plot_mapping(*args, **kwds)

    def plot_image(self, imgdata, multi=False):
        axes = self.axes
        norm = getattr(self.canvas, 'logz', False) and LogNorm() or None
        interp = 'nearest' if not self.imgsmoothing else 'gaussian'
        axes.imshow(imgdata.arr, origin='lower', aspect='equal',
                    interpolation=interp, norm=norm)
        if not multi:
            self.plot_finish(imgdata.xaxis, imgdata.yaxis, imgdata.title)


def plot_mapping(x, y, mapdata, figure=None, axes=None, clear=True, mode=0,
                 colors=None, title=None, dots=True):
    """

    modes: 0 = image
           1 = contour filled
           2 = contour lines
    """
    if figure is None:
        figure = pl.gcf()
    if clear:
        figure.clf()
    if axes is None:
        axes = figure.gca()
    xss, yss, xi, yi, zi = mapdata
    if mode == 0:
        im = axes.imshow(zi.T, origin='lower', aspect='auto',
                         interpolation='nearest',
                         extent=(xi[0][0], xi[-1][-1], yi[0][0], yi[-1][-1]))
    else:
        fcn = axes.contourf if mode == 1 else axes.contour
        kwds = {}
        if colors:
            kwds = {'colors': colors}
        im = fcn(xi, yi, zi, 20,
                 extent=(xi[0][0], xi[-1][-1], yi[0][0], yi[-1][-1]),
                 **kwds)
    axes.set_xlabel(x)
    axes.set_ylabel(y)
    if title is not None:
        axes.set_title(title)
    figure.colorbar(im, ax=axes, fraction=0.05)
    if dots:
        axes.scatter(xss, yss, 0.1)
    return im


def mapping(x, y, runs, minmax=None, mode=0, log=False, dots=True,
            xscale=1, yscale=1, interpolate=100, usemask=True, figure=None,
            clear=True, colors=None, axes=None, title=None):
    from ufit.data.mapping import bin_mapping
    mapdata = bin_mapping(
        x, y, runs, usemask=usemask, log=log, xscale=xscale, yscale=yscale,
        interpolate=interpolate, minmax=minmax)
    return plot_mapping(
        x, y, mapdata, figure=figure, axes=axes, clear=clear, mode=mode,
        dots=dots, colors=colors, title=title)



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2014, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Data loader panel."""

from os import path

from PyQt4.QtCore import pyqtSignature as qtsig, SIGNAL, Qt
from PyQt4.QtGui import QWidget, QFileDialog, QDialogButtonBox, QMessageBox, \
    QMainWindow, QSplitter, QApplication

from ufit.data import data_formats, Loader, ImageData
from ufit.utils import extract_template
from ufit.gui import logger
from ufit.gui.common import loadUi, path_to_str, str_to_path, \
    MPLCanvas, MPLToolbar
from ufit.gui.browse import BrowseWindow
from ufit.gui.common import SettingGroup
from ufit.gui.session import session


class DataLoader(QWidget):

    def __init__(self, parent, plotter, standalone=False):
        QWidget.__init__(self, parent)
        self.logger = logger.getChild('loader')
        self.plotter = plotter
        self.last_data = []
        self.loader = Loader()
        self.createUI(standalone)

        self.sgroup = SettingGroup('main')

        # These will not do anything in standalone mode, but do not hurt.
        self.connect(session, SIGNAL('propsRequested'),
                     self.on_session_propsRequested)
        self.connect(session, SIGNAL('propsUpdated'),
                     self.on_session_propsUpdated)
        self.connect(session, SIGNAL('itemsUpdated'),
                     self.on_session_itemsUpdated)
        self.connect(session, SIGNAL('groupAdded'),
                     self.on_session_itemsUpdated)

        with self.sgroup as settings:
            data_template_path = settings.value('last_data_template', '')
            if data_template_path:
                self.templateEdit.setText(data_template_path)
                self.set_template(data_template_path, 0, silent=True)

    def createUI(self, standalone):
        loadUi(self, 'dataloader.ui')
        self.dataformatBox.addItem('auto')
        for fmt in sorted(data_formats):
            self.dataformatBox.addItem(fmt)

        self.buttonBox.addButton(QDialogButtonBox.Open)
        self.buttonBox.addButton('Preview', QDialogButtonBox.NoRole)

    def on_session_propsRequested(self):
        session.props.template = self.templateEdit.text()

    def on_session_propsUpdated(self):
        if 'template' in session.props:
            self.templateEdit.setText(session.props.template)

    def on_session_itemsUpdated(self):
        # list of groups may have changed
        self.groupBox.clear()
        for group in session.groups:
            self.groupBox.addItem(group.name)

    def on_buttonBox_clicked(self, button):
        role = self.buttonBox.buttonRole(button)
        if role == QDialogButtonBox.RejectRole:
            self.emit(SIGNAL('closeRequest'))
        elif role == QDialogButtonBox.NoRole:  # "preview"
            self.open_data()
        else:  # "open"
            self.open_data(final=True)

    def on_dataformatBox_currentIndexChanged(self, i):
        self.loader.format = str(self.dataformatBox.currentText())

    @qtsig('')
    def on_numorHelpBtn_clicked(self):
        QMessageBox.information(self, 'Numor Help', '''\
The numor string contains file numbers, with the following operators:

, loads multiple files
- loads multiple sequential files
+ merges multiple files
> merges multiple sequential files

For example:

* 10-15,23  loads files 10 through 15 and 23 in 7 separate datasets.
* 10+11,23+24 loads two datasets consisting of files 10 and 11 merged \
into one set, as well as files 23 and 24.
* 10>15+23 merges files 10 through 15 and 23 into one single dataset.
* 10,11,12+13,14 loads four sets.
''')

    def open_browser(self, directory):
        bwin = BrowseWindow(self)
        bwin.show()
        QApplication.processEvents()
        bwin.set_directory(directory)
        bwin.activateWindow()

    @qtsig('')
    def on_browseBtn_clicked(self):
        templ = path_to_str(self.templateEdit.text())
        self.open_browser(path.dirname(templ))

    @qtsig('')
    def on_settemplateBtn_clicked(self):
        previous = path_to_str(self.templateEdit.text())
        if previous:
            startdir = path.dirname(previous)
        else:
            startdir = '.'
        fn = path_to_str(QFileDialog.getOpenFileName(
            self, 'Choose a file', startdir, 'All files (*)'))
        if not fn:
            return
        dtempl, numor = extract_template(fn)
        self.set_template(dtempl, numor)

    def set_template(self, dtempl, numor, silent=True):
        self.templateEdit.setText(str_to_path(dtempl))
        with self.sgroup as settings:
            settings.setValue('last_data_template', dtempl)
        self.loader.template = dtempl
        try:
            cols, xguess, yguess, dyguess, mguess, nmon = \
                self.loader.guess_cols(numor)
        except Exception, e:
            if not silent:
                self.logger.exception('Could not read column names')
                QMessageBox.information(self, 'Error',
                                        'Could not read column names: %s' % e)
            return
        self.xcolBox.clear()
        self.xcolBox.addItem('auto')
        self.xcolBox.setCurrentIndex(0)
        self.ycolBox.clear()
        self.ycolBox.addItem('auto')
        self.ycolBox.setCurrentIndex(0)
        self.dycolBox.clear()
        self.dycolBox.addItem('auto')
        self.dycolBox.addItem('sqrt(Y)')
        self.dycolBox.setCurrentIndex(0)
        self.moncolBox.clear()
        self.moncolBox.addItem('auto')
        self.moncolBox.addItem('none')
        self.moncolBox.setCurrentIndex(0)
        for i, name in enumerate(cols):
            self.xcolBox.addItem(name)
            self.ycolBox.addItem(name)
            self.dycolBox.addItem(name)
            self.moncolBox.addItem(name)
        self.monscaleEdit.setText(str(nmon or 1))
        self.numorsEdit.setText(str(numor))
        self.open_data()

    def open_data(self, final=False):
        try:
            prec = float(self.precisionEdit.text())
        except ValueError:
            QMessageBox.information(self, 'Error', 'Enter a valid precision.')
            return
        floatmerge = self.rbFloatMerge.isChecked()
        xcol = str(self.xcolBox.currentText())
        ycol = str(self.ycolBox.currentText())
        dycol = str(self.dycolBox.currentText())
        mcol = str(self.moncolBox.currentText())
        if mcol == 'none':
            mcol = None
        if dycol == 'sqrt(Y)':
            dycol = None
        try:
            mscale = int(self.monscaleEdit.text())
        except Exception:
            QMessageBox.information(
                self, 'Error', 'Monitor scale must be integer.')
            return
        dtempl = path_to_str(self.templateEdit.text())
        self.loader.template = dtempl
        numors = str(self.numorsEdit.text())
        try:
            datas = self.loader.load_numors(
                numors, prec, xcol, ycol, dycol, mcol, mscale, floatmerge)
        except Exception, e:
            self.logger.exception('Error while loading data file')
            QMessageBox.information(self, 'Error', str(e))
            return
        self.last_data = datas
        if final:
            self.emit(SIGNAL('newDatas'), datas, self.groupBox.currentText())
            self.emit(SIGNAL('closeRequest'))
        else:
            self.plot()

    def initialize(self):
        pass

    def plot(self, limits=True, canvas=None):
        self.plotter.reset()
        xlabels = set()
        ylabels = set()
        titles = set()
        for data in self.last_data:
            xlabels.add(data.xaxis)
            ylabels.add(data.yaxis)
            titles.add(data.title)
            if isinstance(data, ImageData):  # XXX this plots only one
                self.plotter.plot_image(data, multi=True)
                break
            else:
                self.plotter.plot_data(data, multi=True)
        self.plotter.plot_finish(', '.join(xlabels), ', '.join(ylabels),
                                 ', '.join(titles))
        self.plotter.draw()


class DataLoaderMain(QMainWindow):
    def __init__(self, data):
        QMainWindow.__init__(self)
        layout = QSplitter(Qt.Vertical, self)
        self.canvas = MPLCanvas(self)
        self.toolbar = MPLToolbar(self.canvas, self)
        layout.addWidget(self.toolbar)
        layout.addWidget(self.canvas)
        self.dloader = DataLoader(self, self.canvas.plotter, standalone=True)
        self.dloader.groupBox.hide()
        self.dloader.groupBoxLbl.hide()
        self.dloader.groupBoxDesc.hide()
        self.dloader.initialize()
        self.connect(self.dloader, SIGNAL('closeRequest'), self.close)
        layout.addWidget(self.dloader)
        self.setCentralWidget(layout)
        self.setWindowTitle('Data loading')


def start():
    app = QApplication([])
    win = DataLoaderMain()
    win.show()
    app.exec_()
    return win.dloader.last_data
